# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This script is used to launch an experiment on SageMaker. Previously passed arguments on a different
machine are loaded from S3 and the update process is started with these parameters. Arguments
expected are described in renate.benchmark.execute_experiment_job.
"""
import argparse

import renate.defaults as defaults
from renate.benchmark.experimentation import execute_experiment_job


class ExperimentCLI:
    """Entry point to perform an experiment from start to end including evaluation.

    Given a dataset, optionally wrapped in a scenario, a model and a training configuration,
    this script will find the best hyper-parameters per each data chunk, update the model and
    evaluate it on all the provided test data.
    """

    def run(self):
        parser = argparse.ArgumentParser()
        argument_group = parser.add_argument_group("Required Parameters")

        argument_group.add_argument(
            "--config_file",
            type=str,
            required=True,
            help="File containing the definition of model_fn, data_module_fn, config_space_fn and "
            "scheduler_fn.",
        )
        argument_group.add_argument(
            "--experiment_outputs_url",
            type=str,
            required=True,
            help="Location where to store the experimental results and the final model and state.",
        )
        argument_group.add_argument(
            "--num_updates",
            type=int,
            required=True,
            help="How many updates or chunk IDs should be passed to the updater to update the "
            "model.",
        )
        argument_group.add_argument(
            "--mode",
            choices=defaults.SUPPORTED_TUNING_MODE,
            required=True,
            help=f"Declares the type of optimization problem: {defaults.SUPPORTED_TUNING_MODE}",
        )
        argument_group.add_argument(
            "--metric",
            type=str,
            required=True,
            help="Name of metric to optimize.",
        )

        argument_group = parser.add_argument_group("Syne Tune Parameters")
        argument_group.add_argument(
            "--max_time",
            type=float,
            default=None,
            help="Stopping criterion: wall clock time.",
        )
        argument_group.add_argument(
            "--max_num_trials_started",
            type=int,
            default=None,
            help="Stopping criterion: trials started.",
        )
        argument_group.add_argument(
            "--max_num_trials_completed",
            type=int,
            default=None,
            help="Stopping criterion: trials completed.",
        )
        argument_group.add_argument(
            "--max_num_trials_finished",
            type=int,
            default=None,
            help="Stopping criterion: trials finished.",
        )
        argument_group.add_argument(
            "--n_workers",
            type=int,
            default=defaults.N_WORKERS,
            help=f"Number of workers running in parallel. Default: {defaults.N_WORKERS}.",
        )
        parser.add_argument(
            "--working_directory",
            type=str,
            default=defaults.WORKING_DIRECTORY,
            help="Folder used by Renate to store files temporarily. Default: "
            f"{defaults.WORKING_DIRECTORY}.",
        )

        argument_group = parser.add_argument_group("Optional Parameters")
        argument_group.add_argument(
            "--max_epochs",
            type=int,
            default=defaults.MAX_EPOCHS,
            help=f"Maximum number of (finetuning-equiv.) epochs. Default: {defaults.MAX_EPOCHS}",
        )
        argument_group.add_argument(
            "--seed",
            type=int,
            default=defaults.SEED,
            help=f"Seed used for this job. Default: {defaults.SEED}.",
        )
        argument_group.add_argument(
            "--accelerator",
            type=str,
            default=defaults.ACCELERATOR,
            help=f"Accelerator used for this job. Default: {defaults.ACCELERATOR}.",
        )
        argument_group.add_argument(
            "--devices",
            type=int,
            default=defaults.DEVICES,
            help=f"Devices used for this job. Default: {defaults.DEVICES} device.",
        )
        argument_group.add_argument(
            "--backend",
            type=str,
            default="local",
            help=f"Backend to use for the experiment. Can be either {defaults.SUPPORTED_BACKEND}.",
        )
        argument_group = parser.add_argument_group("SageMaker Parameters")
        argument_group.add_argument(
            "--role",
            type=str,
            help="An AWS IAM role (either name or full ARN)",
        )
        argument_group.add_argument(
            "--instance_type",
            type=str,
            default=defaults.INSTANCE_TYPE,
            help="Type of SageMaker instance to use for training, for example, 'ml.c4.xlarge'.",
        )
        argument_group.add_argument(
            "--instance_count",
            type=int,
            default=defaults.INSTANCE_COUNT,
            help="Number of SageMaker instances to use for training.",
        )
        argument_group.add_argument(
            "--instance_max_time",
            type=float,
            default=defaults.INSTANCE_MAX_TIME,
            help="Maximum time in seconds to wait for an instance to be available.",
        )
        argument_group.add_argument(
            "--requirements_file",
            type=str,
            default="./requirements.txt",
            help="File containing the requirements to install Renate.",
        )
        argument_group.add_argument(
            "--job_name",
            type=str,
            default=None,
            help="Name of the SageMaker job.",
        )

        argument_group.add_argument(
            "--deterministic_trainer",
            type=str,
            default=str(defaults.DETERMINISTIC_TRAINER),
            choices=["True", "False"],
            help="When True forces the trainer to be deterministic. Default: "
            f"{defaults.DETERMINISTIC_TRAINER}.",
        )

        args = parser.parse_args()

        execute_experiment_job(
            **vars(args),
        )


if __name__ == "__main__":
    ExperimentCLI().run()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import functools
import logging
from typing import Any, Dict, Optional

import datasets
import torch
import transformers
from datasets import load_dataset

from renate import defaults
from renate.benchmark.datasets.base import DataIncrementalDataModule
from renate.data.data_module import RenateDataModule


class _InputTargetWrapper(torch.utils.data.Dataset):
    """Make a Hugging Face dataset comply with the `(input, target)` format."""

    def __init__(self, dataset, target_column: str = "label"):
        self._dataset = dataset
        self._target_column = target_column

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        item = self._dataset[idx]
        target = item.pop(self._target_column)
        return item, target


class HuggingFaceTextDataModule(RenateDataModule):
    """Data module wrapping Hugging Face text datasets.

    This is a convenience wrapper to expose a Hugging Face dataset as a `RenateDataModule`. Datasets
    will be pre-tokenized and will return `input, target = dataset[i]`, where `input` is a
    dictionary with fields `["input_ids", "attention_mask"]`, and `target` is a tensor.

    We expect the dataset to have a "train" and a "test" split. An additional "validation" split
    will be used if present. Otherwise, a validation set may be split off of the training data
    using the `val_size` argument.

    Args:
        data_path: the path to the folder containing the dataset files.
        tokenizer: Tokenizer to apply to the dataset. See https://huggingface.co/docs/tokenizers/
            for more information on tokenizers.
        dataset_name: Name of the dataset, see https://huggingface.co/datasets. This is a wrapper
            for text datasets only.
        input_column: Name of the column containing the input text.
        target_column: Name of the column containing the target (e.g., class label).
        tokenizer_kwargs: Keyword arguments passed when calling the tokenizer's ``__call__``
           function. Typical options are `max_length`, `padding` and `truncation`.
           See https://huggingface.co/docs/tokenizers/
           for more information on tokenizers. If `None` is passed, this defaults to
           `{"padding": "max_length", max_length: 128, truncation: True}`.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_path: str,
        tokenizer: transformers.PreTrainedTokenizer,
        dataset_name: str = "ag_news",
        input_column: str = "text",
        target_column: str = "label",
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(HuggingFaceTextDataModule, self).__init__(
            data_path=data_path,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name
        self._input_column = input_column
        self._target_column = target_column
        self._tokenizer = tokenizer
        self._tokenizer_kwargs = tokenizer_kwargs or defaults.TOKENIZER_KWARGS

    def prepare_data(self) -> None:
        """Download data."""
        split_names = datasets.get_dataset_split_names(self._dataset_name)
        if "train" not in split_names:
            raise RuntimeError(f"Dataset {self._dataset_name} does not contain a 'train' split.")
        if "test" not in split_names:
            raise RuntimeError(f"Dataset {self._dataset_name} does not contain a 'test' split.")
        self._train_data = datasets.load_dataset(
            self._dataset_name, split="train", cache_dir=self._data_path
        )
        available_columns = list(self._train_data.features)
        if self._input_column not in available_columns:
            raise ValueError(
                f"Input column '{self._input_column}' does not exist in {self._dataset_name}. "
                f"Available columns: {available_columns}."
            )
        if self._target_column not in available_columns:
            raise ValueError(
                f"Target column '{self._target_column}' does not exist in {self._dataset_name}. "
                f"Available columns: {available_columns}."
            )
        self._test_data = datasets.load_dataset(
            self._dataset_name, split="test", cache_dir=self._data_path
        )
        if "validation" in split_names:
            logging.info(f"Using 'validation' split of dataset {self._dataset_name}.")
            self._val_data = datasets.load_dataset(
                self._dataset_name, split="validation", cache_dir=self._data_path
            )
        else:
            logging.info(
                f"No 'validation' split in dataset {self._dataset_name}. Splitting validation data "
                f"from the 'train' split using `val_size={self._val_size}`."
            )
            self._val_data = None

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        self.prepare_data()  # This will use cached datasets if they have already been downloaded.

        def tokenize_fn(batch):
            return self._tokenizer(batch[self._input_column], **self._tokenizer_kwargs)

        columns = ["input_ids", "attention_mask", self._target_column]

        self._train_data = self._train_data.map(tokenize_fn, batched=True)
        self._train_data.set_format(type="torch", columns=columns)
        self._train_data = _InputTargetWrapper(self._train_data, self._target_column)
        self._test_data = self._test_data.map(tokenize_fn, batched=True)
        self._test_data.set_format(type="torch", columns=columns)
        self._test_data = _InputTargetWrapper(self._test_data, self._target_column)
        if self._val_data is not None:
            self._val_data = self._val_data.map(tokenize_fn, batched=True)
            self._val_data.set_format(type="torch", columns=columns)
            self._val_data = _InputTargetWrapper(self._val_data, self._target_column)
        else:
            self._train_data, self._val_data = self._split_train_val_data(self._train_data)


class MultiTextDataModule(DataIncrementalDataModule):
    """
    Inspired by the dataset used in "Episodic Memory in Lifelong Language Learning"
    by d’Autume et al. this is a collection of four different datasets that we call domains:
    AGNews, Yelp, DBPedia and Yahoo Answers.

    The output space if the union of the output space of all the domains.
    The dataset has 33 classes: 4 from AGNews, 5 from Yelp, 14 from DBPedia, and 10 from Yahoo.

    The largest available size for the training set is 115000 and for the test set is 7600.

    Args:
        data_path: The path to the folder where the data files will be downloaded to.
        tokenizer: Tokenizer to apply to the dataset. See https://huggingface.co/docs/tokenizers/
            for more information on tokenizers.
        tokenizer_kwargs: Keyword arguments passed when calling the tokenizer's ``__call__``
           function. Typical options are `max_length`, `padding` and `truncation`.
           See https://huggingface.co/docs/tokenizers/
           for more information on tokenizers. If `None` is passed, this defaults to
           `{"padding": "max_length", max_length: 128, truncation: True}`.
        data_id: The dataset to be used
        train_size: The size of the data stored as training set, must be smaller than 115000.
        test_size: The size of the data stored as test set, must be smaller than 7600.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    _multi_dataset_info = {
        "ag_news": ["text", "label"],
        "yelp_review_full": ["text", "label"],
        "dbpedia_14": ["content", "label"],
        "yahoo_answers_topics": ["question_title", "topic"],
    }
    _label_offset = {
        "ag_news": 0,
        "yelp_review_full": 4,
        "dbpedia_14": 9,
        "yahoo_answers_topics": 23,
    }

    domains = list(_multi_dataset_info)

    def __init__(
        self,
        data_path: str,
        tokenizer: transformers.PreTrainedTokenizer,
        data_id: str,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        train_size: int = 115000,
        test_size: int = 7600,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(data_path=data_path, data_id=data_id, val_size=val_size, seed=seed)

        if train_size > 115000:
            raise ValueError("The `train_size` must be smaller than or equal to 115000")
        self._train_size = train_size

        if test_size > 7600:
            raise ValueError("The `test_size` must be smaller than or equal to 7600")
        self._test_size = test_size

        self._tokenizer = tokenizer
        self._tokenizer_kwargs = tokenizer_kwargs or defaults.TOKENIZER_KWARGS

        if data_id not in self.domains:
            raise ValueError(
                f"The selected domain is not available. Select one among " f"{self.domains}"
            )

        self.data_id = data_id

    def prepare_data(self) -> None:
        """Download dataset."""
        for split in ["train", "test"]:
            load_dataset(self.data_id, split=split, cache_dir=self._data_path)

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        rnd_gen = torch.Generator().manual_seed(self._seed)

        def preprocess(example, text_field_name, label_field_name):
            return {
                **self._tokenizer(example[text_field_name], **self._tokenizer_kwargs),
                "label": example[label_field_name]
                + MultiTextDataModule._label_offset[self.data_id],
            }

        def get_split(split_name):
            dataset = load_dataset(self.data_id, split=split_name, cache_dir=self._data_path)
            # the following is hack needed because the output space of the new dataset is
            # the union of the output spaces of the single datasets
            # HF datasets check for the max label id, and we need to make sure we update that
            # without this change the setup will fail with a value error (label id > max labels)
            new_features = dataset.features.copy()
            new_features[self._multi_dataset_info[self.data_id][1]] = datasets.ClassLabel(
                num_classes=33
            )

            dataset = dataset.cast(new_features)

            if "train" == split_name:
                set_size = self._train_size
            else:
                set_size = self._test_size

            rnd_idx = torch.randint(
                high=len(dataset),
                size=(set_size,),
                generator=rnd_gen,
            ).tolist()
            dataset = dataset.select(indices=rnd_idx)

            dataset = dataset.map(
                functools.partial(
                    preprocess,
                    text_field_name=self._multi_dataset_info[self.data_id][0],
                    label_field_name=self._multi_dataset_info[self.data_id][1],
                ),
                remove_columns=list(dataset.features),
                num_proc=4,
            )

            dataset.set_format(type="torch")

            return _InputTargetWrapper(dataset)

        self._train_data = get_split("train")
        self._train_data, self._val_data = self._split_train_val_data(self._train_data)
        self._test_data = get_split("test")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import numpy as np
import pytest
import torch

from renate.evaluation.metrics.performance_regression_metrics import (
    NegativeFlipImpactMetric,
    NegativeFlipRateMetric,
    PositiveFlipImpactMetric,
    PositiveFlipRateMetric,
)


@pytest.fixture
def input_tensor():
    # nfr = 1 / 4 = 0.25
    # pfr = 1 / 4 = 0.25
    # nfi = 1 / 2 = 0.5
    # pfi = 1 / 2 = 0.5
    new_pred = torch.tensor([1, 3, 0, 3])
    old_pred = torch.tensor([2, 1, 0, 2])
    gt = torch.tensor([3, 3, 0, 2])
    return new_pred, old_pred, gt


@pytest.fixture
def input_numpy_array():
    np_new_pred = np.asarray([1, 3, 0, 3])
    np_old_pred = np.asarray([2, 1, 0, 2])
    np_gt = np.asarray([3, 3, 0, 2])
    return np_new_pred, np_old_pred, np_gt


def test_nfr_metric(input_tensor, input_numpy_array):
    new_pred, old_pred, gt = input_tensor
    np_new_pred, np_old_pred, np_gt = input_numpy_array
    manual_nfr = 0.25

    NFR_metric = NegativeFlipRateMetric()

    correct_old_model = np.equal(np_old_pred, np_gt)
    error_new_model = np.not_equal(np_new_pred, np_gt)
    num_negative_flips = np.logical_and(correct_old_model, error_new_model).sum()

    nfr = num_negative_flips / float(len(gt))

    assert NFR_metric(new_pred, old_pred, gt) == nfr
    assert NFR_metric(new_pred, old_pred, gt) == manual_nfr


def test_pfr_metric(input_tensor, input_numpy_array):
    new_pred, old_pred, gt = input_tensor
    np_new_pred, np_old_pred, np_gt = input_numpy_array
    manual_pfr = 0.25

    PFR_metric = PositiveFlipRateMetric()

    correct_new_model = np.equal(np_new_pred, np_gt)
    error_old_model = np.not_equal(np_old_pred, np_gt)
    num_positive_flips = np.logical_and(correct_new_model, error_old_model).sum()

    pfr = num_positive_flips / float(len(gt))

    assert PFR_metric(new_pred, old_pred, gt) == pfr
    assert PFR_metric(new_pred, old_pred, gt) == manual_pfr


def test_nfi_metric(input_tensor, input_numpy_array):
    new_pred, old_pred, gt = input_tensor
    np_new_pred, np_old_pred, np_gt = input_numpy_array
    manual_nfi = 0.5

    NFI_metric = NegativeFlipImpactMetric()

    correct_old_model = np.equal(np_old_pred, np_gt)
    error_new_model = np.not_equal(np_new_pred, np_gt)
    num_negative_flips = np.logical_and(correct_old_model, error_new_model).sum()

    nfi = num_negative_flips / float(error_new_model.sum())

    assert NFI_metric(new_pred, old_pred, gt) == nfi
    assert NFI_metric(new_pred, old_pred, gt) == manual_nfi


def test_pfi_metric(input_tensor, input_numpy_array):
    new_pred, old_pred, gt = input_tensor
    np_new_pred, np_old_pred, np_gt = input_numpy_array
    manual_pfi = 0.5

    PFI_metric = PositiveFlipImpactMetric()

    correct_new_model = np.equal(np_new_pred, np_gt)
    error_old_model = np.not_equal(np_old_pred, np_gt)
    num_positive_flips = np.logical_and(correct_new_model, error_old_model).sum()

    pfi = num_positive_flips / float(correct_new_model.sum())

    assert PFI_metric(new_pred, old_pred, gt) == pfi
    assert PFI_metric(new_pred, old_pred, gt) == manual_pfi


def test_regression_metric_muti_batch(input_tensor):
    new_pred, old_pred, gt = input_tensor
    new_pred_2 = torch.tensor([3, 1, 1, 3])

    NFR_metric = NegativeFlipRateMetric()
    NFR_metric.update(new_pred, old_pred, gt)
    NFR_metric.update(new_pred_2, old_pred, gt)

    manual_nfr_multi_batch = 0.375

    assert NFR_metric.compute() == manual_nfr_multi_batch


def test_regression_metric_corner_case_empty_input():
    new_pred = torch.tensor([])
    old_pred = torch.tensor([])
    gt = torch.tensor([])

    NFR_metric = NegativeFlipRateMetric()

    with pytest.raises(AssertionError):
        NFR_metric(new_pred, old_pred, gt)


def test_regression_metric_corner_case_unmatch_input():
    new_pred = torch.tensor([1, 2])
    old_pred = torch.tensor([1])
    gt = torch.tensor([1])

    NFR_metric = NegativeFlipRateMetric()

    with pytest.raises(AssertionError):
        NFR_metric(new_pred, old_pred, gt)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.training import run_training_job
from renate.utils.config_spaces import config_space

if __name__ == "__main__":
    # we run the first training job on the MNIST classes [0-4]
    run_training_job(
        config_space=config_space("RD"),  # getting the default search space
        mode="max",
        metric="val_accuracy",
        updater="RD",  # use the RepeatedDistillationModelUpdater
        max_epochs=50,
        chunk_id=0,  # this selects the first chunk of the dataset
        config_file="renate_config.py",
        # location where the new model is saved to
        output_state_url="./state_dump_first_model_rd/",
        backend="local",  # the training job will run on the local machine
        scheduler="asha",
        # using only 5 trials will not give great performance -- just an example
        max_num_trials_finished=5,
    )

    # retrieve the model from `./state_dump_first_model_rd/` if you want
    # do not delete it, we will need it in the remaining of this example

    run_training_job(
        config_space=config_space("RD"),
        mode="max",
        metric="val_accuracy",
        updater="RD",
        max_epochs=50,
        chunk_id=1,  # this time we use the second chunk of the dataset
        config_file="renate_config.py",
        # the output of the first training job is loaded
        input_state_url="./state_dump_first_model_rd/",
        # location where the new model is saved to
        output_state_url="./state_dump_second_model_rd/",
        backend="local",
        scheduler="asha",
        max_num_trials_finished=5,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type

import torch
import torchmetrics
from avalanche.training import Naive
from avalanche.training.plugins import LRSchedulerPlugin
from avalanche.training.supervised.icarl import _ICaRLPlugin
from avalanche.training.templates import BaseSGDTemplate
from syne_tune import Reporter
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import Dataset

from renate import defaults
from renate.models import RenateModule
from renate.updaters.avalanche.learner import (
    AvalancheEWCLearner,
    AvalancheICaRLLearner,
    AvalancheLwFLearner,
    AvalancheReplayLearner,
    ICaRL,
    plugin_by_class,
)
from renate.updaters.avalanche.plugins import (
    RenateCheckpointPlugin,
    RenateFileSystemCheckpointStorage,
)
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.avalanche import AvalancheBenchmarkWrapper, to_avalanche_dataset

logger = logging.getLogger(__name__)

metrics_mapper = {
    "train_loss": "Loss_Epoch/train_phase/train_stream/Task000",
    "train_accuracy": "Top1_Acc_Epoch/train_phase/train_stream/Task000",
    "val_loss": "Loss_Stream/eval_phase/test_stream/Task000",
    "val_accuracy": "Top1_Acc_Stream/eval_phase/test_stream/Task000",
}


class AvalancheModelUpdater(SingleTrainingLoopUpdater):
    _report = Reporter()

    def __init__(self, *args, **kwargs):
        if kwargs.get("mask_unused_classes", False) is True:
            warnings.warn(
                "Avalanche model updaters do not support mask_unused_classes. Ignoring it."
            )
        super().__init__(*args, **kwargs)

    def _load_learner(
        self,
        learner_class: Type[Learner],
        learner_kwargs: Dict[str, Any],
    ) -> BaseSGDTemplate:
        if self._early_stopping_enabled:  # TODO: support it
            raise Exception("Early stopping is not supported yet.")
        logger.warning(
            "Avalanche updaters currently support only accuracy and loss."
        )  # TODO: make use of passed metrics
        self._dummy_learner = learner_class(
            model=self._model,
            **learner_kwargs,
            logged_metrics=self._logged_metrics,
            **self._transforms_kwargs,
        )
        plugins = []
        lr_scheduler_plugin = None
        optimizers_scheduler = self._dummy_learner.configure_optimizers()
        if isinstance(optimizers_scheduler, tuple):
            optimizer, scheduler_config = optimizers_scheduler
            optimizer, scheduler_config = optimizer[0], scheduler_config[0]
            lr_scheduler_plugin = LRSchedulerPlugin(
                scheduler=scheduler_config["scheduler"],
                step_granularity="iteration"
                if scheduler_config["interval"] == "step"
                else scheduler_config["interval"],
            )
            plugins.append(lr_scheduler_plugin)
        else:
            optimizer = optimizers_scheduler
        avalanche_learner = self._load_if_exists(self._input_state_folder)

        checkpoint_plugin = None
        if self._output_state_folder is not None:
            checkpoint_plugin = RenateCheckpointPlugin(
                RenateFileSystemCheckpointStorage(
                    directory=Path(self._output_state_folder),
                ),
            )
            plugins.append(checkpoint_plugin)

        if avalanche_learner is None:
            logger.warning("No updater state available. Updating from scratch.")
            return self._create_avalanche_learner(
                checkpoint_plugin=checkpoint_plugin,
                lr_scheduler_plugin=lr_scheduler_plugin,
                optimizer=optimizer,
            )

        self._dummy_learner.update_settings(
            avalanche_learner=avalanche_learner,
            plugins=plugins,
            optimizer=optimizer,
            max_epochs=self._max_epochs,
            device=self._get_device(),
            eval_every=1,
        )
        return avalanche_learner

    def _get_device(self) -> torch.device:
        """Returns the device according to the chosen accelerator."""
        if self._accelerator == "auto":
            return torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        elif self._accelerator == "gpu":
            return torch.device("cuda", self._devices or 0)
        elif self._accelerator == "tpu":
            raise Exception("Not supported accelerator `TPU`.")
        return torch.device("cpu")

    def _create_avalanche_learner(
        self,
        checkpoint_plugin: RenateCheckpointPlugin,
        lr_scheduler_plugin: Optional[LRSchedulerPlugin],
        optimizer: Optimizer,
    ) -> BaseSGDTemplate:
        """Returns an Avalanche learner based on the arguments passed to the ModelUpdater.

        Args:
            checkpoint_plugin: Plugin to checkpoint regularly.
            lr_scheduler_plugin: Plugin to adapt the learning rate.
            optimizer: PyTorch optimizer object used for training the Avalanche learner.
        """
        plugins = []
        if lr_scheduler_plugin is not None:
            plugins.append(lr_scheduler_plugin)
        if checkpoint_plugin is not None:
            plugins.append(checkpoint_plugin)
        avalanche_learner = self._dummy_learner.create_avalanche_learner(
            optimizer=optimizer,
            train_epochs=self._max_epochs,
            plugins=plugins,
            device=self._get_device(),
            eval_every=1,
        )
        avalanche_learner.is_logged_metric = (
            lambda metric_name: metric_name is None or metric_name in metrics_mapper.keys()
        )
        return avalanche_learner

    @staticmethod
    def _load_if_exists(input_state_folder: Optional[str]) -> Optional[Naive]:
        """Loads the Avalanche strategy if a state exists."""

        if input_state_folder is None:
            return None
        checkpoint_plugin = RenateCheckpointPlugin(
            RenateFileSystemCheckpointStorage(
                directory=Path(input_state_folder),
            ),
        )
        avalanche_learner, _ = checkpoint_plugin.load_checkpoint_if_exists()
        return avalanche_learner

    def update(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> RenateModule:
        val_dataset_exists = val_dataset is not None
        benchmark = self._load_benchmark_if_exists(
            train_dataset, val_dataset, train_dataset_collate_fn, val_dataset_collate_fn
        )
        train_exp = benchmark.train_stream[0]
        self._learner.train(train_exp, eval_streams=[benchmark.test_stream])
        results = self._learner.eval(benchmark.test_stream)
        if isinstance(self._learner, ICaRL):
            class_means = plugin_by_class(_ICaRLPlugin, self._learner.plugins).class_means
            self._model.class_means.data[:, :] = class_means
        if self._output_state_folder is not None:
            Path(self._output_state_folder).mkdir(
                exist_ok=True, parents=True
            )  # TODO: remove when checkpointing is active
            torch.save(self._model.state_dict(), defaults.model_file(self._output_state_folder))
            self._save_avalanche_state(benchmark, val_dataset_exists)
        self._report(
            **{
                metric_name: results[metric_internal_name]
                for metric_name, metric_internal_name in metrics_mapper.items()
                if val_dataset_exists or not metric_name.startswith("val")
            },
            step=self._max_epochs - 1,
            epoch=self._max_epochs,
        )
        return self._model

    def _load_benchmark_if_exists(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
    ) -> AvalancheBenchmarkWrapper:
        avalanche_train_dataset = to_avalanche_dataset(train_dataset, train_dataset_collate_fn)

        avalanche_state = None
        if self._input_state_folder is not None:
            avalanche_state_file = defaults.avalanche_state_file(self._input_state_folder)
            if Path(avalanche_state_file).exists():
                avalanche_state = torch.load(avalanche_state_file)
                if "val_memory_buffer" in avalanche_state:
                    self._dummy_learner._val_memory_buffer.load_state_dict(
                        avalanche_state["val_memory_buffer"]
                    )
                    self._dummy_learner.load(self._input_state_folder)
        if val_dataset is not None:
            self._dummy_learner._val_memory_buffer.update(val_dataset)
            val_memory_dataset = to_avalanche_dataset(
                self._dummy_learner._val_memory_buffer, val_dataset_collate_fn
            )
        else:
            val_memory_dataset = to_avalanche_dataset(train_dataset, val_dataset_collate_fn)

        benchmark = AvalancheBenchmarkWrapper(
            train_dataset=avalanche_train_dataset,
            val_dataset=val_memory_dataset,
            train_transform=self._train_transform,
            train_target_transform=self._train_target_transform,
            test_transform=self._test_transform,
            test_target_transform=self._test_target_transform,
        )
        if avalanche_state is not None:
            benchmark.load_state_dict(avalanche_state)
        benchmark.update_benchmark_properties()
        return benchmark

    def _save_avalanche_state(self, benchmark: AvalancheBenchmarkWrapper, val_dataset_exists: bool):
        state = benchmark.state_dict()
        if val_dataset_exists:
            self._dummy_learner.save(self._output_state_folder)
            state["val_memory_buffer"] = self._dummy_learner._val_memory_buffer.state_dict()
        torch.save(state, defaults.avalanche_state_file(self._output_state_folder))


class ExperienceReplayAvalancheModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        batch_memory_frac: int = defaults.BATCH_MEMORY_FRAC,
        learning_rate_scheduler: Optional[Callable[[Optimizer], _LRScheduler]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "memory_size": memory_size,
            "batch_memory_frac": batch_memory_frac,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=AvalancheReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            mask_unused_classes=mask_unused_classes,
        )


class ElasticWeightConsolidationModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        ewc_lambda: float = defaults.EWC_LAMBDA,
        learning_rate_scheduler: Optional[Callable[[Optimizer], _LRScheduler]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "ewc_lambda": ewc_lambda,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=AvalancheEWCLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            mask_unused_classes=mask_unused_classes,
        )


class LearningWithoutForgettingModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        alpha: float = defaults.LWF_ALPHA,
        temperature: float = defaults.LWF_TEMPERATURE,
        learning_rate_scheduler: Optional[Callable[[Optimizer], _LRScheduler]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "alpha": alpha,
            "temperature": temperature,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=AvalancheLwFLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            mask_unused_classes=mask_unused_classes,
        )


class ICaRLModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        learning_rate_scheduler: Optional[Callable[[Optimizer], _LRScheduler]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=AvalancheICaRLLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import json
import os
from pathlib import Path

import boto3
from syne_tune.backend.sagemaker_backend.sagemaker_utils import get_execution_role

from renate.benchmark.experimentation import execute_experiment_job, experiment_config_file


def load_config(scenario_file, model_file, updater_file, dataset_file):
    cs = {}
    for file in [scenario_file, model_file, updater_file, dataset_file]:
        with open(file) as f:
            cs.update(json.load(f))
    return cs


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--benchmark-file",
        type=str,
        required=True,
        help="Test suite to run.",
    )
    parser.add_argument(
        f"--backend",
        type=str,
        required=True,
        choices=["local", "sagemaker"],
        help="Whether to run locally or on SageMaker.",
    )
    parser.add_argument(
        f"--budget-factor",
        type=float,
        required=True,
        help="Use budget_factor * N many fine-tuning epochs, "
        "where N is the default number of epochs.",
    )
    parser.add_argument(
        f"--job-name",
        type=str,
        required=True,
        help="Name of the job.",
    )
    parser.add_argument(
        f"--num-repetitions",
        type=int,
        default=1,
        help="Number of runs with different seeds.",
    )
    parser.add_argument(
        f"--max-time",
        type=int,
        default=5 * 24 * 3600,
        help="Maximum execution time.",
    )
    parser.add_argument(
        "--use-prior-task-weights",
        action="store_true",
        help="Do not reset model weights (Joint and GDumb only)",
    )

    args = parser.parse_args()
    current_folder = Path(os.path.dirname(__file__))
    configs_folder = current_folder / "experiment_configs"
    benchmark_file = configs_folder / args.benchmark_file
    if not benchmark_file.is_file():
        raise FileNotFoundError(f"Unknown benchmark file '{benchmark_file}'.")
    with open(benchmark_file) as f:
        benchmark_config = json.load(f)
    config_space = load_config(
        os.path.join(configs_folder, "scenarios", benchmark_config["scenario"]),
        os.path.join(configs_folder, "models", benchmark_config["model"]),
        os.path.join(configs_folder, "updaters", benchmark_config["updater"]),
        os.path.join(configs_folder, "datasets", benchmark_config["dataset"]),
    )
    if args.use_prior_task_weights:
        if config_space["updater"] in ["Joint", "GDumb"]:
            config_space["reset"] = False
        else:
            raise ValueError("Please use `use-prior-task-weights` only for Joint or GDumb.")
    config_space["max_epochs"] = int(args.budget_factor * config_space["max_epochs"])
    if "learning_rate_scheduler_step_size" in config_space:
        config_space["learning_rate_scheduler_step_size"] = int(
            args.budget_factor * config_space["learning_rate_scheduler_step_size"]
        )
    if "learning_rate_scheduler_t_max" in config_space:
        config_space["learning_rate_scheduler_t_max"] = int(
            args.budget_factor * config_space["learning_rate_scheduler_t_max"]
        )
    current_folder = Path(os.path.dirname(__file__))

    role = None if args.backend == "local" else get_execution_role()

    for seed in range(args.num_repetitions):
        if args.backend == "local":
            experiment_outputs_url = (
                Path("tmp") / "renate-integration-tests" / args.job_name / str(seed)
            )
            working_directory = str(Path("tmp") / "renate_working_dir")
        else:
            AWS_ACCOUNT_ID = boto3.client("sts").get_caller_identity().get("Account")
            experiment_outputs_url = (
                f"s3://sagemaker-us-west-2-{AWS_ACCOUNT_ID}/renate-domain-incremental/"
                f"{args.job_name}/{seed}"
            )
            working_directory = "/tmp/renate_working_dir"
        execute_experiment_job(
            backend=args.backend,
            config_file=experiment_config_file(),
            config_space=config_space,
            experiment_outputs_url=experiment_outputs_url,
            working_directory=working_directory,
            mode="max",
            metric="val_accuracy",
            num_updates=config_space["num_tasks"],
            role=role,
            instance_type="ml.g4dn.xlarge",
            n_workers=1,
            max_time=args.max_time,
            instance_max_time=args.max_time,
            seed=seed,
            job_name=args.job_name[:36],
            devices=1,
            strategy="ddp",
            save_state=False,
            requirements_file=str(current_folder / "requirements.txt"),
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from renate.benchmark.models.spromptmodel import PromptPool


def test_prompt_pool():
    prompt_size = 7
    embedding_size = 12
    curr_update_id = 3
    pool = PromptPool(
        prompt_size=prompt_size, embedding_size=embedding_size, current_update_id=curr_update_id
    )

    for i in range(curr_update_id):
        assert pool(i).shape == (prompt_size, embedding_size)
        assert pool.get_params(i)[0].shape == (prompt_size, embedding_size)

    pool.increment_task()
    assert len(pool._pool) == curr_update_id + 1



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.benchmark.models import ResNet18
from renate.defaults import TASK_ID


def test_renate_resnet_init():
    pytest.helpers.get_renate_module_resnet(sub_class="resnet18cifar")
    pytest.helpers.get_renate_module_resnet(sub_class="resnet34cifar")
    pytest.helpers.get_renate_module_resnet(sub_class="resnet50cifar")
    pytest.helpers.get_renate_module_resnet(sub_class="resnet18")
    pytest.helpers.get_renate_module_resnet(sub_class="resnet34")
    pytest.helpers.get_renate_module_resnet(sub_class="resnet50")


@pytest.mark.parametrize(
    "sub_class,input_dim",
    [
        ["resnet18cifar", (3, 32, 32)],
        ["resnet34cifar", (3, 32, 32)],
        ["resnet50cifar", (3, 32, 32)],
        ["resnet18", (3, 224, 224)],
        ["resnet34", (3, 224, 224)],
        ["resnet50", (3, 224, 224)],
    ],
)
def test_renate_resnet_fwd(sub_class, input_dim):
    resnet = pytest.helpers.get_renate_module_resnet(sub_class=sub_class)

    x = torch.rand(5, *input_dim)
    y_hat = resnet(x)

    assert y_hat.shape[0] == 5
    assert y_hat.shape[1] == 10


@pytest.mark.parametrize(
    "sub_class, expected_num_params",
    [
        ["resnet18cifar", 60],
        ["resnet34cifar", 108],
        ["resnet50cifar", 159],
        ["resnet18", 60],
        ["resnet34", 108],
        ["resnet50", 159],
    ],
)
def test_renate_resnet_get_params(sub_class, expected_num_params):
    resnet = pytest.helpers.get_renate_module_resnet(sub_class=sub_class)
    resnet.add_task_params(TASK_ID)

    first_task_params = resnet.get_params(TASK_ID)

    resnet.add_task_params("second_task")

    second_task_params = resnet.get_params("second_task")
    assert len(first_task_params) == len(second_task_params)
    assert len(first_task_params) == expected_num_params + 2
    assert len(second_task_params) == expected_num_params + 2
    assert len(list(resnet.parameters())) == expected_num_params + 4
    for i in range(len(first_task_params) - 2):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert torch.equal(first_task_params[i], second_task_params[i])

    for i in range(len(first_task_params) - 2, len(first_task_params)):
        # -2 because the last two parameters are weight and bias of a task specific linear layer
        assert not torch.equal(first_task_params[i], second_task_params[i])


@pytest.mark.parametrize("gray_scale", (True, False), ids=("gray scale", "rgb"))
def test_renate_resnet_gray_scale_parameter(gray_scale):
    """Tests if gray_scale parameter correctly controls number of input channels."""
    expected_in_channels = 1 if gray_scale else 3
    assert ResNet18(gray_scale=gray_scale).get_backbone().conv1.in_channels == expected_in_channels



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Callable, Dict, List, Optional

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from torch.nn import Parameter
from torch.optim import Optimizer

from renate import defaults
from renate.models import RenateModule
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class FineTuningModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "seed": seed,
            "loss_fn": loss_fn,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=Learner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            deterministic_trainer=deterministic_trainer,
            strategy=strategy,
            precision=precision,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Tuple

import torch

from renate.shift.kernels import Kernel


def mmd_gram(
    K: torch.Tensor, z: torch.Tensor, num_permutations: int = 0
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """Maximum mean discrepancy based on a precomputed kernel-Gram matrix.

    This computes the test statistic and (optionally) p-value to conduct an MMD two-sample test to
    decide whether two sets are generated by the same distribution. The inputs are passed implicitly
    in the form of a kernel Gram matrix, evaluated across the union of both sets, and a binary
    vector indicating the assignments of data points to the two sets. I.e., a value of `z[i] = 0`
    indicates that the `i`-th data point belongs to set zero. Optionally, a permutation test is
    carried out and a p-value is returned alongside the raw test statistic.

    MMD tests have been proposed by

    [1] Gretton, A., et al. A kernel two-sample test. JMLR (2012).

    Args:
        K: A tensor containing a kernel-Gram matrix (size `(n, n)`).
        z: A binary vector of length `n` indicating the partition.
        num_permutations: If this is a positive number, a permutation test will be carried out and
            an approximate p-value will be returned.

    Returns:
        A tuple `(t, p)` of two scalar floats, where `t` is the value of the MMD test statistic and
        `p` is the p-value (or `None` if `num_permutations=0`).
    """
    n = K.size(0)
    assert K.size(1) == n
    assert z.size() == (n,)
    inds_0 = torch.where(z == 0)[0]
    inds_1 = torch.where(z == 1)[0]
    n0 = len(inds_0)
    n1 = len(inds_1)

    mmd = (
        K[inds_0][:, inds_0].sum() / (n0 * (n0 - 1))
        + K[inds_1][:, inds_1].sum() / (n1 * (n1 - 1))
        - 2 * K[inds_0][:, inds_1].mean()
    )  # MMD statistic, see Eq. (5) in [1].

    if num_permutations == 0:
        return mmd, None

    # Permutation test: Randomize the assignments z, compute MMD, and count how often we exceed the
    # value obtained with original z.
    cnt = 0
    for _ in range(num_permutations):
        z_ = torch.zeros(n, device=K.device)
        pi = torch.randperm(n, device=K.device)
        z_[pi[:n1]] = 1
        mmd0, _ = mmd_gram(K, z_, num_permutations=0)
        if mmd0 > mmd:
            cnt += 1
    p_val = torch.tensor(cnt / num_permutations)

    return mmd, p_val


def mmd(
    X0: torch.Tensor, X1: torch.Tensor, kernel: Kernel, num_permutations: int = 0
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """Compute MMD between two samples.

    Optionally, return an estimated p-value based on a permutation test.

    Args:
        X0: First sample, shape (n0, dx).
        X1: Second sample, shape (n1, dx).
        kernel: Kernel function to use.
        num_permutations: If this is a positive number, a permutation test will
            be carried out and a p-value will be returned.

    Returns:
        A tuple `(t, p)` where `t` is the value of the MMD test statistic and `p` is the p-value
        (or `None` if `num_permutations=0`).
    """
    X = torch.cat([X0, X1], dim=0)
    z = torch.cat([torch.zeros(X0.size(0)), torch.ones(X1.size(0))], dim=0)
    K = kernel(X, X)
    assert not torch.any(torch.isnan(K))
    return mmd_gram(K, z, num_permutations)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
from collections import defaultdict
from typing import Any, Callable, Dict, Optional, Tuple

import torch
from torch.utils.data import Dataset

from renate import defaults
from renate.data.datasets import IndexedSubsetDataset
from renate.memory.storage import FileTensorStorage
from renate.types import NestedTensors
from renate.utils.pytorch import get_generator

DataDict = Dict[str, torch.Tensor]


class DataBuffer(Dataset):
    """A memory buffer storing data points.

    The buffer functions as a torch dataset, i.e., it implements `__len__` and `__getitem__`.
    Pytorch data loaders can be used to sample from or iterate over the buffer.

    Data can be added to the buffer via `buffer.update(dataset, metadata)`. `dataset` is a
    pytorch dataset expected to return an arbitrary nested `tuple`/`dict` structure containing
    `torch.Tensor`s of _fixed_ size and data type. `metadata` is a dictionary mapping strings to
    tensors for associated metadata. The logic to decide which data points remain in the buffer is
    implemented by different subclasses.

    Extracting an element from the buffer will return a nested tuple of the form
    `data_point, metadata = buffer[i]`, where `data_point` is the raw data point  and `metadata` is
    a dictionary containing associated metadata as well as field `idx` containing the index of the
    data point in the buffer. Additional fields of metadata might be added by some buffering
    methods, e.g., instance weights in coreset methods.

    In addition to passing metadata to the `update` method, one also access and replace the metadata
    in the buffer via the `get_metadata` and `set_metadata` methods.

    Note that, in order to apply transformations, the buffer assumes that the data points are tuples
    of the form `(x, y)`. We apply `transform` to `inputs` and `target_transform` to `y`. Ensure
    that the transforms accept the correct type, e.g., if `x` is a dictionary, `transform` needs to
    operate on a dictionary.
    """

    def __init__(
        self,
        max_size: Optional[int] = None,
        seed: int = defaults.SEED,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        self._max_size = max_size
        self._seed = seed
        self._transform = transform
        self._target_transform = target_transform

        self._rng = get_generator(self._seed)

        self._count = 0
        self._datasets = []
        self._indices = {}
        self._data_point_prototype = None
        self._metadata = {}

    def __len__(self) -> int:
        """Returns the current length of the buffer."""
        return len(self._indices)

    def __getitem__(self, idx: int) -> Tuple[NestedTensors, Dict[str, Any]]:
        """Reads the item at index `idx` of the buffer."""
        i, j = self._indices[idx]
        data = self._datasets[i][j]
        metadata = {key: value[idx] for key, value in self._metadata.items()}
        if self._transform is None and self._target_transform is None:
            return data, metadata
        else:
            inputs, targets = data
            if self._transform is not None:
                inputs = self._transform(inputs)
            if self._target_transform is not None:
                targets = self._target_transform(targets)
            return (inputs, targets), metadata

    def update(self, dataset: Dataset, metadata: Optional[Dict] = None) -> None:
        """Updates the buffer with a new dataset."""
        metadata = metadata or {}
        self._check_metadata_internal_consistency(metadata, expected_length=len(dataset))
        if not len(self):
            self._data_point_prototype = copy.deepcopy(dataset[0])
            self._add_metadata_like(metadata)
        else:
            self._check_metadata_compatibility(metadata)
        assignments = self._update(dataset, metadata)
        self._check_assignments_are_contiguous(assignments)

        # Perform assignments.
        d = len(self._datasets)
        self._datasets.append(dataset)
        for buffer_idx, dataset_idx in assignments.items():
            self._indices[buffer_idx] = (d, dataset_idx)
            for key in self._metadata.keys():
                self._metadata[key][buffer_idx] = metadata[key][dataset_idx]

    def _update(self, dataset: Dataset, metadata: Dict) -> Dict[int, int]:
        """Returns the updates.

        This method is used to implement different buffering methods. Given a dataset, it returns a
        dictionary of `assignments` indicating the updates to be performed to the buffer.

        Args:
            dataset: The dataset used to update the buffer.

        Returns:
            assignments: A dictionary mapping int to int. The presence of the pair `(i, j)` means
                that the `i`-th buffer slot will be filled with the `j`-th element of `dataset`.
        """
        raise NotImplementedError()

    def get_metadata(self, key: str) -> Dict[str, torch.Tensor]:
        return self._metadata[key][: len(self)]

    def set_metadata(self, key: str, values: torch.Tensor) -> None:
        if values.size(0) != len(self):
            raise ValueError()
        if key not in self._metadata:
            self._add_metadata_like({key: values})
        self._metadata[key][: len(self)] = values.cpu()

    def state_dict(self) -> Dict:
        return {
            "buffer_class_name": self.__class__.__name__,
            "max_size": self._max_size,
            "seed": self._seed,
            "count": self._count,
            "indices": self._indices,
            "data_point_prototype": self._data_point_prototype,
            "metadata": self._metadata,
        }

    def load_state_dict(self, state_dict: Dict) -> None:
        if self.__class__.__name__ != state_dict["buffer_class_name"]:
            raise RuntimeError(
                f"Buffer of class {self.__class__} was used to load a state dict created by class "
                f"{state_dict['buffer_class_name']}."
            )
        self._max_size = state_dict["max_size"]
        self._seed = state_dict["seed"]
        self._count = state_dict["count"]
        self._indices = state_dict["indices"]
        self._data_point_prototype = state_dict["data_point_prototype"]
        self._metadata = state_dict["metadata"]
        self._rng = get_generator(self._seed)

    def save(self, target_dir: str) -> None:
        if not len(self):
            return

        transforms = self._transform, self._target_transform
        self._transform, self._target_transform = None, None
        storage = FileTensorStorage(target_dir)
        storage.dump_dataset(IndexedSubsetDataset(self, [0]))
        self._datasets = [storage]
        self._indices = {i: (0, i) for i in range(len(self))}
        self._transform, self._target_transform = transforms

    def load(self, source_dir: str) -> None:
        if not len(self):
            return

        storage = FileTensorStorage(source_dir)
        self._datasets = [storage]

    def _add_metadata_like(self, metadata: DataDict):
        for key, value in metadata.items():
            self._metadata[key] = torch.empty(
                (self._max_size, *value.size()[1:]), dtype=value.dtype, device="cpu"
            )

    def _check_assignments_are_contiguous(self, assignments):
        # The subclass _update method is expected to only return assignments that populate the
        # buffer contiguously. We check this here. The invariant we expect to hold is that, at any
        # time, the indices 0:len(buffer) are filled.
        current_length = len(self)
        new_idx = [i for i in assignments.keys() if i > current_length]
        if new_idx:
            assert max(new_idx) == current_length + len(new_idx)

    @staticmethod
    def _check_metadata_internal_consistency(metadata: DataDict, expected_length: int):
        """Checks metadata for internal consistency and an expected length."""
        for key in metadata:
            if not isinstance(metadata[key], torch.Tensor):
                raise TypeError(
                    f"Expected tensors in `metadata`, found {type(metadata[key])} at key {key}."
                )
            if not metadata[key].size(0) == expected_length:
                raise ValueError(
                    f"Tensors in metadata dictionary need to be of size {expected_length} (size "
                    f"of the associated dataset) in dimension 0. Found size {metadata[key].size()} "
                    f"at key {key}."
                )

    def _check_metadata_compatibility(self, metadata: DataDict) -> None:
        """Verifies that passed metadata is compatible with internal metadata."""
        if not len(self):
            return
        if set(self._metadata.keys()) != set(metadata.keys()):
            raise KeyError(
                f"Keys of provided metadata {list(metadata.keys())} do not match those present in ",
                f"the buffer {list(self._metadata.keys())}.",
            )
        for key in metadata:
            if metadata[key].size()[1:] != self._metadata[key].size()[1:]:
                raise ValueError()
            if metadata[key].dtype != self._metadata[key].dtype:
                raise TypeError(
                    f"Provided metadata at key {key} is of type {metadata[key].dtype}. This does "
                    f"not match type {self._metadata[key].dtype} already present in the buffer."
                )


class ReservoirBuffer(DataBuffer):
    """A buffer implementing reservoir sampling."""

    def _update(self, dataset: Dataset, metadata: Dict) -> Dict[int, int]:
        assignments = {}
        for i in range(len(dataset)):
            if self._count < self._max_size:
                assignments[self._count] = i
            else:
                rand = torch.randint(low=0, high=self._count, size=(), generator=self._rng).item()
                if rand < self._max_size:
                    assignments[rand] = i
            self._count += 1
        return assignments


class SlidingWindowBuffer(DataBuffer):
    """A buffer implementing a sliding window."""

    def _update(self, dataset: Dataset, metadata) -> Dict[int, int]:
        assignments = {}
        for i in range(len(dataset)):
            assignments[self._count % self._max_size] = i
            self._count += 1
        return assignments


class GreedyClassBalancingBuffer(DataBuffer):
    """A buffer implementing a greedy class-balancing approach."""

    def __init__(
        self,
        max_size: Optional[int] = None,
        seed: int = defaults.SEED,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        super().__init__(max_size, seed, transform, target_transform)
        self._indices_by_class = defaultdict(list)

    def state_dict(self) -> Dict:
        state_dict = super().state_dict()
        state_dict["indices_by_class"] = self._indices_by_class
        return state_dict

    def load_state_dict(self, state_dict: Dict) -> None:
        super().load_state_dict(state_dict)
        self._indices_by_class = defaultdict(list)
        self._indices_by_class.update(state_dict["indices_by_class"])

    def _get_largest_class(self):
        largest_classes = []
        max_length = 0
        for y, indices in self._indices_by_class.items():
            if len(indices) > max_length:
                largest_classes = [y]
                max_length = len(indices)
            if len(indices) == max_length:
                largest_classes.append(y)
        rand = torch.randint(low=0, high=len(largest_classes), size=(1,), generator=self._rng)
        largest_class = largest_classes[rand.item()]
        return largest_class, max_length

    def _update(self, dataset: Dataset, metadata: Dict) -> Dict[int, int]:
        assignments = {}
        for i in range(len(dataset)):
            y = int(dataset[i][1])
            if self._count < self._max_size:
                assignments[self._count] = i
                self._indices_by_class[y].append(self._count)
            else:
                largest_class, max_length = self._get_largest_class()
                rand = torch.randint(low=0, high=max_length, size=(1,), generator=self._rng).item()
                idx_to_replace = self._indices_by_class[largest_class].pop(rand)
                assignments[idx_to_replace] = i
                self._indices_by_class[y].append(idx_to_replace)
            self._count += 1
        return assignments


class InfiniteBuffer(DataBuffer):
    """A data buffer that stores _all_ incoming data."""

    def __init__(
        self,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        super().__init__(max_size=1, transform=transform, target_transform=target_transform)

    def _update(self, dataset: Dataset, metadata: Dict) -> Dict[int, int]:
        # Resize metadata dictionary.
        if self._count + len(dataset) > self._max_size:
            while self._count + len(dataset) > self._max_size:
                self._max_size *= 2
            current_metadata = {key: self.get_metadata(key).clone() for key in self._metadata}
            self._add_metadata_like(current_metadata)
            for key, values in current_metadata.items():
                self.set_metadata(key, values)

        assignments = {self._count + i: i for i in range(len(dataset))}
        self._count += len(dataset)
        return assignments



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

from renate.utils.file import copy_to_uri, move_to_uri


def create_file(path, content):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        f.write(content)


@pytest.mark.parametrize(
    "file_content1_source_dir, file_content2_source_dir, file_content1_destination_dir,"
    "file_content2_destination_dir",
    [["test1_1", "test1_1", "test1_2", "test2_2"]],
)
def test_move_to_uri_locally_directory(
    tmpdir,
    file_content1_source_dir,
    file_content2_source_dir,
    file_content1_destination_dir,
    file_content2_destination_dir,
):
    """Test for moving files from a local directory to another local directory.

    The files should be moved from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.
    """

    create_file(os.path.join(tmpdir, "source_dir", "file1.txt"), file_content1_source_dir)
    create_file(os.path.join(tmpdir, "source_dir", "file2.txt"), file_content2_source_dir)

    create_file(os.path.join(tmpdir, "destination_dir", "file1.txt"), file_content1_destination_dir)
    create_file(os.path.join(tmpdir, "destination_dir", "file3.txt"), file_content2_destination_dir)

    move_to_uri(os.path.join(tmpdir, "source_dir"), os.path.join(tmpdir, "destination_dir"))

    with open(os.path.join(tmpdir, "destination_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file3.txt"), "r") as f:
        assert f.read() == file_content2_destination_dir

    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file1.txt"))
    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file2.txt"))


def test_move_to_uri_locally_file(tmpdir):
    """Test for moving a file to another local directory.

    The file should be moved from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.
    """

    file_content = "content"
    create_file(os.path.join(tmpdir, "source_dir", "file.txt"), file_content)

    move_to_uri(
        os.path.join(tmpdir, "source_dir", "file.txt"), os.path.join(tmpdir, "destination_dir")
    )
    with open(os.path.join(tmpdir, "destination_dir", "file.txt"), "r") as f:
        assert f.read() == file_content
    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file.txt"))

    file_content = "content2"
    create_file(os.path.join(tmpdir, "source_dir", "file.txt"), file_content)
    move_to_uri(
        os.path.join(tmpdir, "source_dir", "file.txt"), os.path.join(tmpdir, "destination_dir")
    )
    with open(os.path.join(tmpdir, "destination_dir", "file.txt"), "r") as f:
        assert f.read() == file_content
    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file.txt"))


@pytest.mark.parametrize(
    "file_content1_source_dir, file_content2_source_dir, file_content1_destination_dir, "
    "file_content2_destination_dir",
    [["test1_1", "test1_1", "test1_2", "test2_2"]],
)
def test_copy_to_uri_locally_directory(
    tmpdir,
    file_content1_source_dir,
    file_content2_source_dir,
    file_content1_destination_dir,
    file_content2_destination_dir,
):
    """Test for copying files from a local directory to another local directory.

    The files should be copied from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.

    The source directory should not be changed.
    """

    create_file(os.path.join(tmpdir, "source_dir", "file1.txt"), file_content1_source_dir)
    create_file(os.path.join(tmpdir, "source_dir", "file2.txt"), file_content2_source_dir)

    create_file(os.path.join(tmpdir, "destination_dir", "file1.txt"), file_content1_destination_dir)
    create_file(os.path.join(tmpdir, "destination_dir", "file3.txt"), file_content2_destination_dir)

    copy_to_uri(os.path.join(tmpdir, "source_dir"), os.path.join(tmpdir, "destination_dir"))

    with open(os.path.join(tmpdir, "destination_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file3.txt"), "r") as f:
        assert f.read() == file_content2_destination_dir

    with open(os.path.join(tmpdir, "source_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "source_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir


def test_copy_to_uri_locally_file(tmpdir):
    """Test for copying a file from a local directory to another local directory.

    The file should be copied from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.

    The source directory should not be changed.
    """
    file_content = "content"
    create_file(os.path.join(tmpdir, "source_dir", "file.txt"), file_content)

    copy_to_uri(
        os.path.join(tmpdir, "source_dir", "file.txt"), os.path.join(tmpdir, "destination_dir")
    )
    with open(os.path.join(tmpdir, "destination_dir", "file.txt"), "r") as f:
        assert f.read() == file_content
    with open(os.path.join(tmpdir, "source_dir", "file.txt"), "r") as f:
        assert f.read() == file_content



# Copyright 2020 The PyTorch Lightning team and Microsoft Corporation. All rights reserved.
# Modifications: Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved
# SPDX-License-Identifier: Apache-2.0

import os
import pickle as pkl
from pathlib import Path
from typing import Any, Dict, Optional, Union

import torch
from deepspeed.utils.zero_to_fp32 import (
    get_fp32_state_dict_from_zero_checkpoint,
    get_model_state_file,
    get_optim_files,
)

CPU_DEVICE = torch.device("cpu")


def ds_checkpoint_dir(checkpoint_dir: Union[str, Path], tag: Optional[str] = None) -> str:
    if tag is None:
        latest_path = os.path.join(checkpoint_dir, "latest")
        if os.path.isfile(latest_path):
            with open(latest_path) as fd:
                tag = fd.read().strip()
        else:
            raise ValueError(f"Unable to find 'latest' file at {latest_path}")

    directory = os.path.join(checkpoint_dir, tag)

    if not os.path.isdir(directory):
        raise FileNotFoundError(f"Directory '{ds_checkpoint_dir}' doesn't exist")
    return directory


# modified
def search_key(state: Dict[str, Any], substring: str) -> str:
    """This function looks for a substring in keys of dict and returns the full key that
    is the first match."""
    for k in state.keys():
        if substring in k:
            return k


# Modified script from
# https://github.com/Lightning-AI/lightning/blob/1.8.6/src/pytorch_lightning/utilities/deepspeed.py
# which is modified from
# https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/utils/zero_to_fp32.py
def convert_zero_checkpoint_to_fp32_state_dict(
    checkpoint_dir: Union[str, Path], tag: Optional[str] = None
) -> Dict[str, Any]:
    """Convert ZeRO 2 or 3 checkpoint into a single fp32 consolidated ``state_dict`` file
    that can be loaded with ``torch.load(file)`` + ``load_state_dict()`` and used for training
    without DeepSpeed.  Additionally the script has been modified to ensure we keep the
    lightning state inside the state dict for being able to run
    ``LightningModule.load_from_checkpoint('...')``.
    Modification to this version include the explicit handling of the _extra_state
    element of state dict. Deepspeed's and Lightning get-fp-32... functions only collate
    trainable parameters.

    Args:
        checkpoint_dir: path to the desired checkpoint folder.
            (one that contains the tag-folder, like ``global_step14``)
        tag: checkpoint tag used as a unique identifier for checkpoint. If not provided will
            attempt to load tag in the file named ``latest`` in the checkpoint folder,
            e.g., ``global_step14``.
    """

    state_dict = get_fp32_state_dict_from_zero_checkpoint(checkpoint_dir, tag)
    # additional logic to ensure we keep the lightning state dict as well from rank 0.
    deepspeed_states = [
        "module",
        "optimizer",
        "lr_scheduler",
        "csr_tensor_module_names",
        "skipped_steps",
        "global_steps",
        "dp_world_size",
        "mp_world_size",
    ]
    checkpoint_dir = ds_checkpoint_dir(checkpoint_dir)
    optim_files = get_optim_files(checkpoint_dir)
    optim_state = torch.load(optim_files[0], map_location=CPU_DEVICE)
    zero_stage = optim_state["optimizer_state_dict"]["zero_stage"]
    model_file = get_model_state_file(checkpoint_dir, zero_stage)
    client_state = torch.load(model_file, map_location=CPU_DEVICE)
    # Assign extra_state by searching for which key it is
    extra_key = search_key(client_state["module"], "extra_state")
    extra_state = client_state["module"][extra_key]
    state_dict[extra_key] = extra_state
    # End of modifications
    client_state = {
        key: value for key, value in client_state.items() if key not in deepspeed_states
    }
    # State dict keys will include reference to wrapper _LightningModuleWrapperBase
    # Delete `module` prefix before saving.
    state_dict = {k.partition("module.")[2]: state_dict[k] for k in state_dict.keys()}
    client_state["state_dict"] = state_dict

    return client_state


def convert_to_tensor(obj):
    """This function converts a pickleable object to a torch tensor. This is only to
    aid saving with Deepspeed."""
    return torch.as_tensor(list(pkl.dumps(obj)))


def recover_object_from_tensor(tensor):
    """This function converts a tensor to a byte stream that is passed through pickle
    to recover the underlying object. For usage with Deepspeed"""
    return pkl.loads(bytes(tensor.tolist()))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from syne_tune.config_space import loguniform, uniform
from syne_tune.optimizer.baselines import ASHA
from syne_tune.optimizer.schedulers import FIFOScheduler
from syne_tune.optimizer.schedulers.transfer_learning import RUSHScheduler

from renate import defaults
from renate.training.training import (
    RENATE_CONFIG_COLUMNS,
    _create_scheduler,
    _get_transfer_learning_task_evaluations,
    _load_tuning_history,
    _merge_tuning_history,
    _verify_validation_set_for_hpo_and_checkpointing,
    run_training_job,
)
from renate.utils.syne_tune import is_syne_tune_config_space

config_file = str(Path(__file__).parent.parent / "renate_config_files" / "config.py")
config_file_custom_optimizer = str(
    Path(__file__).parent.parent / "renate_config_files" / "config_custom_optimizer.py"
)


@pytest.mark.parametrize(
    "num_chunks, val_size, raises, fixed_search_space, scheduler, configuration_file",
    [
        (2, 0.9, False, False, "rush", config_file),
        (1, 0.0, True, False, "rush", config_file),
        (2, 0.9, False, True, None, config_file),
        (2, 0.0, False, True, None, config_file),
        (1, 0.0, False, True, None, config_file_custom_optimizer),
    ],
    ids=[
        "transfer-hpo-with-val",
        "transfer-hpo-without-val-raises-exception",
        "training-single-config-with-val",
        "training-single-config-without-val",
        "training-single-config-without-val-custom-optimizer",
    ],
)
@pytest.mark.parametrize("updater", ("ER", "Avalanche-iCaRL"))
def test_run_training_job(
    tmpdir, num_chunks, val_size, raises, fixed_search_space, scheduler, updater, configuration_file
):
    """Simply running tuning job to check if anything fails.

    Case 1: Standard HPO setup with transfer learning in second step.
    Case 2: HPO without validation set fails.
    Case 3: Training of single configuration with validation set.
    Case 4: Training of single configuration without validation set.
    Case 5: Training of single configuration without validation set using custom optimizer and
        learning rate scheduler..
    """
    state_url = None
    tmpdir = str(tmpdir)
    for _ in range(num_chunks):

        def execute_job():
            config_space = {"val_size": val_size}
            if configuration_file == config_file:
                config_space["learning_rate"] = (
                    0.1 if fixed_search_space else loguniform(10e-5, 0.1)
                )
            run_training_job(
                updater=updater,
                max_epochs=5,
                config_file=configuration_file,
                input_state_url=state_url,
                output_state_url=tmpdir,
                backend="local",
                mode="max",
                config_space=config_space,
                metric="val_accuracy",
                max_time=35,
                scheduler=scheduler,
            )

        if raises:
            with pytest.raises(
                AssertionError, match="Provide a validation set to optimize hyperparameters."
            ):
                execute_job()
            return
        execute_job()
        state_url = tmpdir


@pytest.mark.parametrize(
    "data_old, data_new, expected_data_first_row, expected_data_second_row",
    [
        (
            [[0, 0.001, 0.01]],
            [[0.002, 0.8, 0.02]],
            [0, 0.001, np.nan, 0.01],
            [1, 0.002, 0.8, 0.02],
        ),
        (None, [[0.002, 0.8, 0.02]], [0, 0.002, 0.8, 0.02], None),
    ],
)
def test_merge_tuning_history(
    data_old, data_new, expected_data_first_row, expected_data_second_row
):
    """Test whether HPO results are merged correctly.

    Testing two cases:
        1. Old results exist and new ones are added.
        2. No old results exist.
    """
    results_old = None
    if data_old is not None:
        results_old = pd.DataFrame(data_old, columns=["update_id", "val_loss", "config_lr"])
    results_new = pd.DataFrame(data_new, columns=["val_loss", "val_accuracy", "config_lr"])
    merged_results = _merge_tuning_history(
        new_tuning_results=results_new, old_tuning_results=results_old
    )
    np.testing.assert_array_equal(
        merged_results[["update_id", "val_loss", "val_accuracy", "config_lr"]].iloc[0],
        expected_data_first_row,
    )
    if expected_data_second_row is not None:
        np.testing.assert_array_equal(
            merged_results[["update_id", "val_loss", "val_accuracy", "config_lr"]].iloc[1],
            expected_data_second_row,
        )


@pytest.mark.parametrize("val_size", [0.9, 0.0])
@pytest.mark.parametrize("tune_hyperparameters", [True, False])
def test_verify_validation_set_for_hpo_and_checkpointing(tmpdir, val_size, tune_hyperparameters):
    """Check if misconfigurations are spotted and config_space is updated correctly.

    If tune_hyperparameters is `True` (hyperparameter optimization is enabled), a validation set
    must exist.
    If a validation set exists, the `config_space` must be changed such that the right metric and
    mode for checkpointing and hyperparameter optimization is used.
    """
    config_space = {"val_size": val_size}
    expected_metric = "val_accuracy"
    expected_mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "max"

    def verify_validation_set():
        return _verify_validation_set_for_hpo_and_checkpointing(
            config_space=config_space,
            config_file=config_file,
            tune_hyperparameters=tune_hyperparameters,
            metric=expected_metric,
            mode=expected_mode,
            working_directory=tmpdir,
        )

    if tune_hyperparameters and val_size == 0:
        with pytest.raises(AssertionError):
            verify_validation_set()
    else:
        metric, mode = verify_validation_set()
        if val_size > 0:
            assert config_space.get("metric") == expected_metric
            assert config_space.get("mode") == expected_mode
            assert metric == expected_metric
            assert mode == expected_mode
        else:
            assert metric == "train_loss"
            assert mode == "min"


def test_is_syne_tune_config_space():
    """Function should return True if any entry in the `config_space` is a Syne Tune search
    space.
    """
    assert is_syne_tune_config_space({"a": uniform(0, 1), "b": 1})
    assert not is_syne_tune_config_space({"a": 0.5, "b": 1})


def _get_transfer_learning_task_evaluations_input():
    """Helper function which returns the input required for the unit tests below."""
    hyperparameter_names = ["config_lr"]
    objectives_names = ["train_loss", "val_loss"]

    tuning_results = pd.DataFrame(
        [
            [[0.3, 0.2], [0.33, 0.23], 0.1, "irrelevant"],
            [[0.2, 0.1, 0.01], [0.23, 0.13, 0.03], 0.01, "irrelevant"],
            [[0.4], [0.5], 0.4, "irrelevant"],
            [[0.5], [0.6], 0.5, "irrelevant"],
            [[0.51], [0.61], 0.5, "irrelevant"],  # duplicate hyperparameter
            [[0.5], [np.nan], 0.3, "irrelevant"],  # missing metric
            [[0.5], [0.6], np.nan, "irrelevant"],  # missing hyperparameter
        ],
        columns=objectives_names + hyperparameter_names + ["irrelevant_columns"],
    )
    config_space = {"lr": uniform(0.001, 1)}
    for renate_config in RENATE_CONFIG_COLUMNS:
        if renate_config != "state_url":
            tuning_results[f"config_{renate_config}"] = f"{renate_config}_old"
        config_space[renate_config] = f"{renate_config}_new"

    return {
        "tuning_results": tuning_results,
        "config_space": config_space,
        "metric": objectives_names[1],
        "max_epochs": 3,
    }


def test_get_transfer_learning_task_evaluations():
    """Case that should successfully return a result.

    Function must drop duplicate hyperparameters, rows where metric or hyperparameter are missing.
    Only the relevant metric is kept. Renate configurations are replaced with the current ones
    (e.g. working space).
    """
    input_dict = _get_transfer_learning_task_evaluations_input()
    expected_objectives_evaluations = np.array(
        [
            [[[0.33], [0.23], [np.nan]]],
            [[[0.23], [0.13], [0.03]]],
            [[[0.5], [np.nan], [np.nan]]],
            [[[0.6], [np.nan], [np.nan]]],
        ]
    )
    expected_hyperparameters = [0.1, 0.01, 0.4, 0.5]
    task_evaluations = _get_transfer_learning_task_evaluations(
        tuning_results=input_dict["tuning_results"],
        config_space=input_dict["config_space"],
        metric=input_dict["metric"],
        max_epochs=input_dict["max_epochs"],
    )
    assert task_evaluations is not None
    assert task_evaluations.objectives_names == [input_dict["metric"]]
    assert task_evaluations.objectives_evaluations.shape == (4, 1, 3, 1)
    np.testing.assert_array_equal(
        task_evaluations.objectives_evaluations, expected_objectives_evaluations
    )
    assert list(task_evaluations.hyperparameters.columns) == ["lr"] + RENATE_CONFIG_COLUMNS
    assert task_evaluations.hyperparameters["lr"].to_list() == expected_hyperparameters
    for hyperparameter in task_evaluations.hyperparameters:
        if hyperparameter != "lr":
            assert (
                task_evaluations.hyperparameters[hyperparameter] == f"{hyperparameter}_new"
            ).all()


@pytest.mark.parametrize(
    "updated_input_dict",
    [
        {"config_space": {"unknown_hyperparameter": uniform(0, 1), "lr": uniform(0, 1)}},
        {"metric": "val_unknown_metric"},
        {"tuning_results": pd.DataFrame(columns=["train_loss", "val_loss", "config_lr"])},
        {
            "tuning_results": _get_transfer_learning_task_evaluations_input()[
                "tuning_results"
            ].iloc[5:6]
        },
        {
            "tuning_results": _get_transfer_learning_task_evaluations_input()[
                "tuning_results"
            ].iloc[6:7]
        },
    ],
)
def test_get_transfer_learning_task_evaluations_failure_cases(updated_input_dict):
    """Cases in which the function should return `None` since metadata cannot be used.

    Case 1: The `config_space` contains a new hyperparameter.
    Case 2: No evaluations for the given `metric` exist.
    Case 3: No evaluations exist.
    Case 4: Metric column is all `np.nan`.
    Case 5: Any hyperparameter column contains only `np.nan`.
    """
    input_dict = _get_transfer_learning_task_evaluations_input()
    input_dict.update(updated_input_dict)
    task_evaluations = _get_transfer_learning_task_evaluations(
        tuning_results=input_dict["tuning_results"],
        config_space=input_dict["config_space"],
        metric=input_dict["metric"],
        max_epochs=input_dict["max_epochs"],
    )
    assert task_evaluations is None


def test_load_tuning_history_with_no_tuning_history_should_return_empty_dict(tmpdir):
    assert _load_tuning_history(tmpdir, {}, "val_loss") == {}


def _get_tuning_results():
    return pd.DataFrame(
        [
            [0, 0, 1, 0.3, 0.4, 0.1, np.nan],
            [0, 0, 2, 0.2, 0.3, 0.1, np.nan],
            [0, 0, 3, 0.1, 0.2, 0.1, np.nan],
            [1, 0, 1, 0.3, 0.4, 0.1, 0.8],
            [1, 1, 1, 0.2, 0.3, 0.05, 0.9],
            [1, 0, 2, 0.25, 0.45, 0.1, 0.8],
            [1, 2, 1, 0.1, 0.2, 0.05, 0.8],
        ],
        columns=[
            "update_id",
            "trial_id",
            "epoch",
            "train_loss",
            "val_loss",
            "config_lr",
            "config_momentum",
        ],
    )


@pytest.mark.parametrize("hyperparameter, num_task_evaluations", [("lr", 2), ("momentum", 1)])
def test_load_tuning_history(tmpdir, hyperparameter, num_task_evaluations):
    """Check if tuning results can be successfully converted to a list of
    TransferLearningTaskEvaluations.

    Case 1: Evaluations for hyperparameter `lr` is available for both updates. Return both.
    Case 2: Evaluations for hyperparameter `momentum` is only available for the second update.
        Drop first data.
    """
    tuning_results = _get_tuning_results()
    config_space = {hyperparameter: uniform(0.001, 1)}
    tuning_results.to_csv(defaults.hpo_file(tmpdir), index=False)
    task_evaluations = _load_tuning_history(tmpdir, config_space, "val_loss")
    assert len(task_evaluations) == num_task_evaluations
    assert None not in task_evaluations.values()


@pytest.mark.parametrize("use_dir", [True, False])
def test_load_tuning_history_when_no_previous_history_exists(tmpdir, use_dir):
    """Check if function returns empty dict in case no previous history exists.

    Case 1: state_url is None. Return empty dict.
    Case 2: state_url is a path which does not contain a tuning history. Return empty dict.
    """
    state_url = None
    if use_dir:
        state_url = tmpdir
    task_evaluations = _load_tuning_history(state_url, {"lr": 0.01}, "val_loss")
    assert task_evaluations == {}


@pytest.mark.parametrize(
    "tuning_results_exist", [True, False], ids=["with_tuning_results", "without_tuning_results"]
)
@pytest.mark.parametrize(
    "scheduler,scheduler_kwargs",
    [
        ("random", None),
        ("asha", None),
        ("rush", None),
        (FIFOScheduler, None),
        (ASHA, {"max_resource_attr": "max_epochs", "resource_attr": "epoch"}),
        (RUSHScheduler, {"max_resource_attr": "max_epochs", "resource_attr": "epoch"}),
    ],
    ids=["str_random", "str_asha", "str_rush", "class_random", "class_asha", "class_rush"],
)
def test_create_scheduler(tmpdir, scheduler, scheduler_kwargs, tuning_results_exist):
    """Test various ways of creating a scheduler with and without previous tuning results.

    Cases:
        1. Check if schedulers can be created by using a string only.
        2. Custom Syne Tune schedulers can be created by passing class and kwargs.
        3. If any of the above methods works with and without previous tuning results.
    """
    tuning_results = _get_tuning_results()
    state_url = None
    if tuning_results_exist:
        state_url = tmpdir
        tuning_results.to_csv(defaults.hpo_file(state_url), index=False)
    _create_scheduler(
        scheduler=scheduler,
        config_space={"lr": 0.01, "max_epochs": 3},
        metric="val_loss",
        mode="min",
        seed=0,
        scheduler_kwargs=scheduler_kwargs,
        input_state_url=state_url,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from .buffer import (
    DataBuffer,
    GreedyClassBalancingBuffer,
    InfiniteBuffer,
    ReservoirBuffer,
    SlidingWindowBuffer,
)

__all__ = [
    "DataBuffer",
    "GreedyClassBalancingBuffer",
    "InfiniteBuffer",
    "ReservoirBuffer",
    "SlidingWindowBuffer",
]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
import shutil
from typing import Callable, Dict, Literal

import pytest
import torch
from pytorch_lightning.loggers import TensorBoardLogger
from renate import defaults
from renate.benchmark.models import (
    MultiLayerPerceptron,
    ResNet18,
    ResNet18CIFAR,
    ResNet34,
    ResNet34CIFAR,
    ResNet50,
    ResNet50CIFAR,
    VisionTransformerB16,
    VisionTransformerB32,
    VisionTransformerCIFAR,
    VisionTransformerH14,
    VisionTransformerL16,
    VisionTransformerL32,
)
from renate.models.renate_module import RenateModule
from renate.updaters.avalanche.learner import (
    AvalancheEWCLearner,
    AvalancheICaRLLearner,
    AvalancheLwFLearner,
    AvalancheReplayLearner,
)
from renate.updaters.avalanche.model_updater import AvalancheModelUpdater
from renate.updaters.experimental.er import ExperienceReplayLearner
from renate.updaters.experimental.gdumb import GDumbLearner
from renate.updaters.experimental.joint import JointLearner
from renate.updaters.experimental.l2p import LearningToPromptLearner, LearningToPromptReplayLearner
from renate.updaters.experimental.offline_er import OfflineExperienceReplayLearner
from renate.updaters.experimental.repeated_distill import RepeatedDistillationLearner
from renate.updaters.learner import Learner, ReplayLearner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.optimizer import create_partial_optimizer

pytest_plugins = ["helpers_namespace"]


def pytest_addoption(parser):
    parser.addoption(
        "--runslow",
        action="store_true",
        default=False,
        help="Run slow tests e.g. testing data modules.",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        return
    skip_slow = pytest.mark.skip(reason="Need --runslow option to run.")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


LEARNER_KWARGS = {
    ExperienceReplayLearner: {
        "memory_size": 30,
        "batch_memory_frac": 0.4,
        "batch_size": 50,
        "seed": 1,
    },
    Learner: {"batch_size": 10, "seed": 42},
    LearningToPromptLearner: {"batch_size": 10, "seed": 42, "prompt_sim_loss_weight": 1},
    LearningToPromptReplayLearner: {
        "batch_size": 10,
        "seed": 42,
        "prompt_sim_loss_weight": 1,
        "loss_weight_new_data": 0.5,
        "memory_size": 30,
        "batch_memory_frac": 0.3,
    },
    GDumbLearner: {
        "batch_size": 10,
        "seed": 42,
        "memory_size": 30,
    },
    JointLearner: {"batch_size": 10, "seed": 3},
    RepeatedDistillationLearner: {"batch_size": 10, "seed": 42, "memory_size": 30},
    OfflineExperienceReplayLearner: {
        "memory_size": 30,
        "batch_memory_frac": 0.4,
        "loss_weight_new_data": 0.5,
        "batch_size": 50,
        "seed": 1,
    },
}
AVALANCHE_LEARNER_KWARGS = {
    AvalancheReplayLearner: {
        "memory_size": 30,
        "batch_memory_frac": 0.4,
        "batch_size": 50,
        "seed": 1,
    },
    AvalancheEWCLearner: {
        "ewc_lambda": 0.1,
        "batch_size": 50,
        "seed": 1,
    },
    AvalancheLwFLearner: {
        "alpha": 0.1,
        "temperature": 2,
        "batch_size": 50,
        "seed": 1,
    },
    AvalancheICaRLLearner: {
        "memory_size": 30,
        "batch_size": 50,
        "seed": 1,
    },
}
AVALANCHE_LEARNER_HYPERPARAMETER_UPDATES = {
    AvalancheEWCLearner: {"ewc_lambda": 0.3},
    AvalancheLwFLearner: {"alpha": 0.2, "temperature": 3},
    AvalancheICaRLLearner: {},
    AvalancheReplayLearner: {},
}
LEARNERS = list(LEARNER_KWARGS)
AVALANCHE_LEARNERS = list(AVALANCHE_LEARNER_KWARGS)
LEARNERS_USING_SIMPLE_UPDATER = [
    ExperienceReplayLearner,
    Learner,
    GDumbLearner,
    JointLearner,
    OfflineExperienceReplayLearner,
]
L2P_LEARNERS = [
    LearningToPromptLearner,
    LearningToPromptReplayLearner,
]

SAMPLE_CLASSIFICATION_RESULTS = {
    "accuracy": [
        [0.9362000226974487, 0.6093000173568726, 0.3325999975204468],
        [0.8284000158309937, 0.9506999850273132, 0.3382999897003174],
        [0.4377000033855438, 0.48260000348091125, 0.9438999891281128],
    ],
    "accuracy_init": [[0.2, 0.1, 0.09]],
}

TEST_WORKING_DIRECTORY = "./test_renate_working_dir/"
TEST_LOGGER = TensorBoardLogger
TEST_LOGGER_KWARGS = {"save_dir": TEST_WORKING_DIRECTORY, "version": 1, "name": "lightning_logs"}


@pytest.helpers.register
def get_renate_module_mlp(
    num_inputs, num_outputs, num_hidden_layers, hidden_size, add_icarl_class_means=False
) -> RenateModule:
    return MultiLayerPerceptron(
        num_inputs,
        num_outputs,
        num_hidden_layers,
        hidden_size,
        add_icarl_class_means=add_icarl_class_means,
    )


@pytest.helpers.register
def get_loss_fn(reduction="none") -> torch.nn.Module:
    return torch.nn.CrossEntropyLoss(reduction=reduction)


@pytest.helpers.register
def get_renate_module_resnet(sub_class="resnet18cifar", **kwargs) -> RenateModule:
    kwargs["add_icarl_class_means"] = False
    if sub_class == "resnet18cifar":
        return ResNet18CIFAR(**kwargs)
    elif sub_class == "resnet34cifar":
        return ResNet34CIFAR(**kwargs)
    elif sub_class == "resnet50cifar":
        return ResNet50CIFAR(**kwargs)
    elif sub_class == "resnet18":
        return ResNet18(**kwargs)
    elif sub_class == "resnet34":
        return ResNet34(**kwargs)
    elif sub_class == "resnet50":
        return ResNet50(**kwargs)
    else:
        raise ValueError("Invalid ResNet called.")


@pytest.helpers.register
def get_renate_module_vision_transformer(
    sub_class="visiontransformerb16", **kwargs
) -> RenateModule:
    kwargs["add_icarl_class_means"] = False
    if sub_class == "visiontransformercifar":
        return VisionTransformerCIFAR(**kwargs)
    elif sub_class == "visiontransformerb16":
        return VisionTransformerB16(**kwargs)
    elif sub_class == "visiontransformerb32":
        return VisionTransformerB32(**kwargs)
    elif sub_class == "visiontransformerl16":
        return VisionTransformerL16(**kwargs)
    elif sub_class == "visiontransformerl32":
        return VisionTransformerL32(**kwargs)
    elif sub_class == "visiontransformerh14":
        return VisionTransformerH14(**kwargs)
    else:
        raise ValueError("Invalid Vision Transformer called.")


@pytest.helpers.register
def get_renate_vision_module(model, sub_class="resnet18cifar", **kwargs):
    if model == "resnet":
        return get_renate_module_resnet(sub_class, **kwargs)
    elif model == "visiontransformer":
        return get_renate_module_vision_transformer(sub_class, **kwargs)
    else:
        raise ValueError("Invalid vision model called.")


@pytest.helpers.register
def get_renate_module_mlp_and_data(
    num_inputs,
    num_outputs,
    num_hidden_layers,
    hidden_size,
    train_num_samples,
    test_num_samples,
    val_num_samples=0,
    add_icarl_class_means=False,
):
    model = get_renate_module_mlp(
        num_inputs=num_inputs,
        num_outputs=num_outputs,
        hidden_size=hidden_size,
        num_hidden_layers=num_hidden_layers,
        add_icarl_class_means=add_icarl_class_means,
    )
    train_dataset = torch.utils.data.TensorDataset(
        torch.rand(train_num_samples, num_inputs),
        torch.randint(num_outputs, (train_num_samples,)),
    )
    test_data = torch.rand(test_num_samples, num_inputs)

    if val_num_samples > 0:
        val_dataset = torch.utils.data.TensorDataset(
            torch.rand(val_num_samples, num_inputs),
            torch.randint(num_outputs, (val_num_samples,)),
        )
        return model, train_dataset, val_dataset

    return model, train_dataset, test_data


@pytest.helpers.register
def get_renate_vision_module_and_data(
    input_size,
    num_outputs,
    train_num_samples,
    test_num_samples,
    model="resnet",
    sub_class="reduced18",
    **kwargs,
):
    model = get_renate_vision_module(model, sub_class, **kwargs)
    train_dataset = torch.utils.data.TensorDataset(
        torch.rand(train_num_samples, *input_size),
        torch.randint(num_outputs, (train_num_samples,)),
    )
    test_data = torch.rand(test_num_samples, *input_size)
    return model, train_dataset, test_data


@pytest.helpers.register
def get_simple_updater(
    model,
    partial_optimizer=None,
    input_state_folder=None,
    output_state_folder=None,
    learner_class=ExperienceReplayLearner,
    learner_kwargs=None,
    max_epochs=5,
    train_transform=None,
    train_target_transform=None,
    test_transform=None,
    test_target_transform=None,
    buffer_transform=None,
    buffer_target_transform=None,
    early_stopping_enabled=False,
    metric=None,
    deterministic_trainer=False,
):
    if learner_kwargs is None:
        learner_kwargs = {"memory_size": 10}
    transforms_kwargs = {
        "train_transform": train_transform,
        "train_target_transform": train_target_transform,
        "test_transform": test_transform,
        "test_target_transform": test_target_transform,
    }
    if issubclass(learner_class, ReplayLearner):
        transforms_kwargs["buffer_transform"] = buffer_transform
        transforms_kwargs["buffer_target_transform"] = buffer_target_transform
    return SingleTrainingLoopUpdater(
        model=model,
        loss_fn=get_loss_fn(),
        optimizer=partial_optimizer or get_partial_optimizer(),
        learner_class=learner_class,
        learner_kwargs=learner_kwargs,
        input_state_folder=input_state_folder,
        output_state_folder=output_state_folder,
        max_epochs=max_epochs,
        accelerator="cpu",
        logger=TEST_LOGGER(**TEST_LOGGER_KWARGS),
        early_stopping_enabled=early_stopping_enabled,
        metric=metric,
        deterministic_trainer=deterministic_trainer,
        **transforms_kwargs,
    )


@pytest.helpers.register
def get_avalanche_updater(
    model,
    input_state_folder=None,
    output_state_folder=None,
    learner_class=AvalancheReplayLearner,
    learner_kwargs=None,
    max_epochs=5,
    train_transform=None,
    train_target_transform=None,
    test_transform=None,
    test_target_transform=None,
    early_stopping_enabled=False,
    metric=None,
):
    if learner_kwargs is None:
        learner_kwargs = {"memory_size": 10}
    transforms_kwargs = {
        "train_transform": train_transform,
        "train_target_transform": train_target_transform,
        "test_transform": test_transform,
        "test_target_transform": test_target_transform,
    }
    return AvalancheModelUpdater(
        model=model,
        loss_fn=get_loss_fn("mean"),
        optimizer=get_partial_optimizer(),
        learner_class=learner_class,
        learner_kwargs=learner_kwargs,
        input_state_folder=input_state_folder,
        output_state_folder=output_state_folder,
        max_epochs=max_epochs,
        accelerator="cpu",
        early_stopping_enabled=early_stopping_enabled,
        metric=metric,
        **transforms_kwargs,
    )


@pytest.helpers.register
def get_partial_optimizer(
    optimizer: Literal["Adam", "SGD"] = defaults.OPTIMIZER,
    lr: float = defaults.LEARNING_RATE,
    momentum: float = defaults.MOMENTUM,
    weight_decay: float = defaults.WEIGHT_DECAY,
):
    return create_partial_optimizer(
        optimizer=optimizer, lr=lr, momentum=momentum, weight_decay=weight_decay
    )


@pytest.helpers.register
def check_learner_transforms(learner: Learner, expected_transforms: Dict[str, Callable]):
    """Checks if the learner transforms match to expected ones.

    Args:
        learner: The learner which transforms will be checked.
        expected_transforms: Dictionairy mapping from transform name to transform. These are the
            expected transforms for the learner.
    """
    assert learner._train_transform is expected_transforms.get(
        "train_transform"
    ) and learner._train_target_transform is expected_transforms.get("train_target_transform")
    if isinstance(learner, ReplayLearner):
        assert learner._memory_buffer._transform is expected_transforms.get(
            "buffer_transform"
        ) and learner._memory_buffer._target_transform is expected_transforms.get(
            "buffer_target_transform"
        )


def pytest_sessionstart(session):
    if not os.path.exists(TEST_WORKING_DIRECTORY):
        os.mkdir(TEST_WORKING_DIRECTORY)


def pytest_sessionfinish(session, exitstatus):
    shutil.rmtree(TEST_WORKING_DIRECTORY)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pytest
import torch

from renate import defaults
from renate.updaters.experimental.er import (
    CLSExperienceReplayLearner,
    DarkExperienceReplayLearner,
    ExperienceReplayLearner,
    PooledOutputDistillationExperienceReplayLearner,
    SuperExperienceReplayLearner,
)


def get_model_and_dataset():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    dataset = torch.utils.data.TensorDataset(
        torch.rand((100, 10)),
        torch.randint(10, (100,)),
    )
    return model, dataset


@pytest.mark.parametrize(
    "batch_size,memory_size,batch_memory_frac",
    [[20, 10, 0.5], [30, 10, 0.34], [20, 100, 0.5], [10, 30, 0.1], [100, 10, 0.03]],
)
def test_er_overall_memory_size_after_update(batch_size, memory_size, batch_memory_frac):
    memory_batch_size = int(batch_memory_frac * batch_size)
    model, dataset = get_model_and_dataset()
    learner_kwargs = {
        "memory_size": memory_size,
        "batch_memory_frac": batch_memory_frac,
        "batch_size": batch_size,
    }
    model_updater = pytest.helpers.get_simple_updater(
        model=model,
        partial_optimizer=pytest.helpers.get_partial_optimizer(),
        learner_class=ExperienceReplayLearner,
        learner_kwargs=learner_kwargs,
        max_epochs=1,
    )
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    memory, _ = next(iter(model_updater._learner._memory_loader))
    x, y = memory
    assert x.shape[0] == memory_batch_size and y.shape[0] == memory_batch_size
    assert len(model_updater._learner._memory_buffer) == memory_size


def test_er_validation_buffer(tmpdir):
    model, dataset_train, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    datasets_val = []
    state_folder = None
    next_state_folder = defaults.input_state_folder(tmpdir)
    for i in range(3):
        dataset_val = torch.utils.data.TensorDataset(
            torch.rand((100, 10)),
            torch.randint(10, (100,)),
        )
        model_updater = pytest.helpers.get_simple_updater(
            model, input_state_folder=state_folder, output_state_folder=next_state_folder
        )
        model_updater.update(
            train_dataset=dataset_train, val_dataset=dataset_val, task_id=defaults.TASK_ID
        )
        datasets_val.append(dataset_val)
        state_folder = next_state_folder

    model_updater = pytest.helpers.get_simple_updater(
        model, input_state_folder=state_folder, output_state_folder=next_state_folder
    )
    for i in range(3):
        for j in range(100):
            assert torch.allclose(
                datasets_val[i][j][0],
                model_updater._learner._val_memory_buffer[i * 100 + j][0][0],
                rtol=1e-3,
            )


def validate_common_args(model_updater, learner_kwargs):
    memory_batch_size = int(learner_kwargs["batch_memory_frac"] * learner_kwargs["batch_size"])
    batch_size = learner_kwargs["batch_size"] - memory_batch_size
    assert model_updater._learner._batch_size == batch_size
    assert model_updater._learner._memory_batch_size == memory_batch_size


def validate_cls_er(model_updater, learner_kwargs):
    validate_common_args(model_updater, learner_kwargs)
    assert model_updater._learner._components["memory_loss"].weight == learner_kwargs["alpha"]
    assert model_updater._learner._components["cls_loss"].weight == learner_kwargs["beta"]
    assert (
        model_updater._learner._components["cls_loss"]._stable_model_update_weight
        == learner_kwargs["stable_model_update_weight"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._plastic_model_update_weight
        == learner_kwargs["plastic_model_update_weight"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._stable_model_update_probability
        == learner_kwargs["stable_model_update_probability"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._plastic_model_update_probability
        == learner_kwargs["plastic_model_update_probability"]
    )


def validate_dark_er(model_updater, learner_kwargs):
    validate_common_args(model_updater, learner_kwargs)
    assert model_updater._learner._components["memory_loss"].weight == learner_kwargs["beta"]
    assert model_updater._learner._components["mse_loss"].weight == learner_kwargs["alpha"]


def validate_pod_er(model_updater, learner_kwargs):
    validate_common_args(model_updater, learner_kwargs)
    assert model_updater._learner._components["pod_loss"].weight == learner_kwargs["alpha"]
    assert (
        model_updater._learner._components["pod_loss"]._distillation_type
        == learner_kwargs["distillation_type"]
    )
    assert model_updater._learner._components["pod_loss"]._normalize == learner_kwargs["normalize"]


def validate_super_er(model_updater, learner_kwargs):
    validate_common_args(model_updater, learner_kwargs)
    assert model_updater._learner._components["memory_loss"].weight == learner_kwargs["der_beta"]
    assert model_updater._learner._components["mse_loss"].weight == learner_kwargs["der_alpha"]
    assert model_updater._learner._components["cls_loss"].weight == learner_kwargs["cls_alpha"]
    assert (
        model_updater._learner._components["cls_loss"]._stable_model_update_weight
        == learner_kwargs["cls_stable_model_update_weight"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._plastic_model_update_weight
        == learner_kwargs["cls_plastic_model_update_weight"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._stable_model_update_probability
        == learner_kwargs["cls_stable_model_update_probability"]
    )
    assert (
        model_updater._learner._components["cls_loss"]._plastic_model_update_probability
        == learner_kwargs["cls_plastic_model_update_probability"]
    )
    assert model_updater._learner._components["pod_loss"].weight == learner_kwargs["pod_alpha"]
    assert (
        model_updater._learner._components["pod_loss"]._distillation_type
        == learner_kwargs["pod_distillation_type"]
    )
    assert (
        model_updater._learner._components["pod_loss"]._normalize == learner_kwargs["pod_normalize"]
    )
    assert (
        model_updater._learner._components["shrink_perturb"]._shrink_factor
        == learner_kwargs["sp_shrink_factor"]
    )
    assert model_updater._learner._components["shrink_perturb"]._sigma == learner_kwargs["sp_sigma"]


@pytest.mark.parametrize(
    "learner_class,validate_fct,learner_kwargs1,learner_kwargs2",
    [
        (
            CLSExperienceReplayLearner,
            validate_cls_er,
            {
                "memory_size": 30,
                "batch_memory_frac": 0.4,
                "batch_size": 50,
                "seed": 1,
                "alpha": 0.123,
                "beta": 2,
                "stable_model_update_weight": 0.2,
                "plastic_model_update_weight": 0.1,
                "stable_model_update_probability": 0.3,
                "plastic_model_update_probability": 0.4,
            },
            {
                "memory_size": 30,
                "batch_memory_frac": 0.1,
                "batch_size": 100,
                "seed": 1,
                "alpha": 2.3,
                "beta": 3,
                "stable_model_update_weight": 0.6,
                "plastic_model_update_weight": 0.5,
                "stable_model_update_probability": 0.7,
                "plastic_model_update_probability": 0.8,
            },
        ),
        (
            DarkExperienceReplayLearner,
            validate_dark_er,
            {
                "memory_size": 30,
                "batch_memory_frac": 0.4,
                "batch_size": 50,
                "seed": 1,
                "alpha": 0.123,
                "beta": 2,
            },
            {
                "memory_size": 30,
                "batch_memory_frac": 0.1,
                "batch_size": 100,
                "seed": 1,
                "alpha": 2.3,
                "beta": 3,
            },
        ),
        (
            PooledOutputDistillationExperienceReplayLearner,
            validate_pod_er,
            {
                "memory_size": 30,
                "batch_memory_frac": 0.4,
                "batch_size": 50,
                "seed": 1,
                "alpha": 0.123,
                "distillation_type": "spatial",
                "normalize": True,
            },
            {
                "memory_size": 30,
                "batch_memory_frac": 0.1,
                "batch_size": 100,
                "seed": 1,
                "alpha": 0.123,
                "distillation_type": "channel",
                "normalize": False,
            },
        ),
        (
            SuperExperienceReplayLearner,
            validate_super_er,
            {
                "memory_size": 30,
                "batch_memory_frac": 0.4,
                "batch_size": 50,
                "seed": 1,
                "der_alpha": 0.123,
                "der_beta": 2,
                "sp_shrink_factor": 0.33,
                "sp_sigma": 0.11,
                "cls_alpha": 2.3,
                "cls_stable_model_update_weight": 0.6,
                "cls_plastic_model_update_weight": 0.5,
                "cls_stable_model_update_probability": 0.7,
                "cls_plastic_model_update_probability": 0.8,
                "pod_alpha": 0.13,
                "pod_distillation_type": "spatial",
                "pod_normalize": True,
            },
            {
                "memory_size": 30,
                "batch_memory_frac": 0.1,
                "batch_size": 100,
                "seed": 1,
                "der_alpha": 2.3,
                "der_beta": 3,
                "sp_shrink_factor": 0.66,
                "sp_sigma": 0.22,
                "cls_alpha": 2.3,
                "cls_stable_model_update_weight": 0.6,
                "cls_plastic_model_update_weight": 0.5,
                "cls_stable_model_update_probability": 0.7,
                "cls_plastic_model_update_probability": 0.8,
                "pod_alpha": 0.123,
                "pod_distillation_type": "channel",
                "pod_normalize": False,
            },
        ),
    ],
)
def test_saving_and_loading_of_er_methods(
    tmpdir, learner_class, validate_fct, learner_kwargs1, learner_kwargs2
):
    """ER saving and loading test.

    The ER methods partially have custom saving and loading functions (CLS-ER). Furthermore, the
    components used to be Modules that effectively did not allow for changing hyperparameter
    settings.
    """
    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=1,
        train_num_samples=10,
        test_num_samples=5,
    )
    state_url = defaults.input_state_folder(tmpdir)

    model_updater = pytest.helpers.get_simple_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=learner_kwargs1,
        output_state_folder=state_url,
        max_epochs=2,
    )

    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    validate_fct(model_updater, learner_kwargs1)
    model_updater = pytest.helpers.get_simple_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=learner_kwargs2,
        input_state_folder=state_url,
        max_epochs=2,
    )
    validate_fct(model_updater, learner_kwargs2)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from renate.benchmark.models.mlp import MultiLayerPerceptron
from renate.benchmark.models.resnet import (
    ResNet18,
    ResNet18CIFAR,
    ResNet34,
    ResNet34CIFAR,
    ResNet50,
    ResNet50CIFAR,
)
from renate.benchmark.models.l2p import LearningToPromptTransformer, PromptedTransformer
from renate.benchmark.models.spromptmodel import SPromptTransformer
from renate.benchmark.models.vision_transformer import (
    VisionTransformerB16,
    VisionTransformerB32,
    VisionTransformerCIFAR,
    VisionTransformerH14,
    VisionTransformerL16,
    VisionTransformerL32,
)

__all__ = [
    "MultiLayerPerceptron",
    "ResNet18",
    "ResNet18CIFAR",
    "ResNet34",
    "ResNet34CIFAR",
    "ResNet50",
    "ResNet50CIFAR",
    "LearningToPromptTransformer",
    "PromptedTransformer",
    "SPromptTransformer",
    "VisionTransformerB16",
    "VisionTransformerB32",
    "VisionTransformerCIFAR",
    "VisionTransformerH14",
    "VisionTransformerL16",
    "VisionTransformerL32",
]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torch.nn.functional as F

from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner_components.component import Component


class WeightedLossComponent(Component, ABC):
    """The abstract class implementing a weighted loss function.

    This is an abstract class from which each other loss should inherit from.
    """

    def _verify_attributes(self) -> None:
        """Verify if attributes have valid values."""
        super()._verify_attributes()
        assert self.weight >= 0, "Weight must be larger than 0."

    def loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        if self.weight == 0:
            return torch.tensor(0.0)
        return self._loss(
            outputs_memory=outputs_memory,
            batch_memory=batch_memory,
            intermediate_representation_memory=intermediate_representation_memory,
        )

    @abstractmethod
    def _loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        pass


class WeightedCustomLossComponent(WeightedLossComponent):
    """Adds a (weighted) user-provided custom loss contribution.

    Args:
        loss_fn: The loss function to apply.
        weight: A scaling coefficient which should scale the loss which gets returned.
        sample_new_memory_batch: Whether a new batch of data should be sampled from the memory
            buffer when the loss is calculated.
    """

    def __init__(self, loss_fn: Callable, weight: float, sample_new_memory_batch: bool) -> None:
        super().__init__(weight=weight, sample_new_memory_batch=sample_new_memory_batch)
        self._loss_fn = loss_fn

    def _loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        """Returns user-provided loss evaluated on memory batch."""
        (_, targets_memory), _ = batch_memory
        return self.weight * self._loss_fn(outputs_memory, targets_memory)


class WeightedMeanSquaredErrorLossComponent(WeightedLossComponent):
    """Mean squared error between the current and previous logits computed with respect to the
    memory sample.
    """

    def _loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        """Mean-squared error between current and previous logits on memory."""
        logits = outputs_memory
        _, meta_data = batch_memory
        previous_logits = meta_data["outputs"]
        return self.weight * F.mse_loss(logits, previous_logits, reduction="mean")


class WeightedPooledOutputDistillationLossComponent(WeightedLossComponent):
    """Pooled output feature distillation with respect to intermediate network features.

    As described in: Douillard, Arthur, et al. "Podnet: Pooled outputs distillation for small-tasks
    incremental learning." European Conference on Computer Vision. Springer, Cham, 2020.

    Given the intermediate representations collected at different parts of the network, minimise
    their Euclidean distance with respect to the cached representation. There are different
    `distillation_type`s trading-off plasticity and stability of the resultant representations.
    `normalize` enables the user to normalize the resultant feature representations to ensure that
    they are less affected by their magnitude.

    Args:
        weight: Scaling coefficient which scales the loss with respect to all intermediate
            representations.
        sample_new_memory_batch: Whether a new batch of data should be sampled from the memory
            buffer when the loss is calculated.
        distillation_type: Which distillation type to apply with respect to all intermediate
            representations.
        normalize: Whether to normalize both the current and cached features before computing the
            Frobenius norm.
    """

    def __init__(
        self,
        weight: float,
        sample_new_memory_batch: bool,
        distillation_type: str = "spatial",
        normalize: bool = True,
    ) -> None:
        self._distillation_type = distillation_type
        super().__init__(weight=weight, sample_new_memory_batch=sample_new_memory_batch)
        self._normalize = normalize

    def _verify_attributes(self) -> None:
        """Verify if attributes have valid values."""
        super()._verify_attributes()
        if self._distillation_type not in ["pixel", "channel", "width", "height", "gap", "spatial"]:
            raise ValueError(f"Invalid distillation type: {self._distillation_type}")

    def _sum_reshape(self, x: torch.Tensor, dim: int) -> torch.Tensor:
        """Sum the tensor according to specific dimension and reshape."""
        batch_size = x.shape[0]
        return x.sum(dim=dim).reshape(batch_size, -1)

    def _pod(self, features: torch.Tensor, features_memory: torch.Tensor) -> torch.Tensor:
        """Pooled output distillation with respect to intermediate and cached intermediate features.

        Args:
            features: Current intermediate features.
            features_memory: Cached intermediate features.
        """
        if features.shape != features_memory.shape:
            raise ValueError(
                "The shape of the features and the cached features should be the same: "
                f"{features.shape}, and: {features_memory.shape}"
            )

        features = features.pow(2)
        features_memory = features_memory.pow(2)

        if self._distillation_type == "channel":
            features, features_memory = self._sum_reshape(features, 1), self._sum_reshape(
                features_memory, 1
            )
        elif self._distillation_type == "width":
            features, features_memory = self._sum_reshape(features, 2), self._sum_reshape(
                features_memory, 2
            )
        elif self._distillation_type == "height":
            features, features_memory = self._sum_reshape(features, 3), self._sum_reshape(
                features_memory, 3
            )
        elif self._distillation_type == "gap":
            features = F.adaptive_avg_pool2d(features, (1, 1))[..., 0, 0]
            features_memory = F.adaptive_avg_pool2d(features_memory, (1, 1))[..., 0, 0]
        elif self._distillation_type == "spatial":
            features_h, features_memory_h = self._sum_reshape(features, 3), self._sum_reshape(
                features_memory, 3
            )
            features_w, features_memory_w = self._sum_reshape(features, 2), self._sum_reshape(
                features_memory, 2
            )
            features = torch.cat([features_h, features_w], dim=-1)
            features_memory = torch.cat([features_memory_h, features_memory_w], dim=-1)

        if self._normalize:
            features = F.normalize(features, dim=1, p=2)
            features_memory = F.normalize(features_memory, dim=1, p=2)

        return torch.frobenius_norm(features - features_memory, dim=-1).mean(dim=0)

    def _loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: List[torch.Tensor],
    ) -> torch.Tensor:
        """Compute the pooled output with respect to current and cached intermediate outputs from
        memory.
        """
        loss = torch.tensor(0.0)
        _, meta_data = batch_memory
        for n in range(len(intermediate_representation_memory)):
            features = intermediate_representation_memory[n]
            features_memory = meta_data[f"intermediate_representation_{n}"]
            loss += self._pod(features, features_memory)
        return (self.weight * loss) / len(intermediate_representation_memory)


class WeightedCLSLossComponent(WeightedLossComponent):
    """Complementary Learning Systems Based Experience Replay.

    Arani, Elahe, Fahad Sarfraz, and Bahram Zonooz.
    "Learning fast, learning slow: A general continual learning method based on complementary
    learning system." arXiv preprint arXiv:2201.12604 (2022).

    The implementation follows the Algorithm 1 in the respective paper. The complete `Learner`
    implementing this loss, is the `CLSExperienceReplayLearner`.

    Args:
        weight: A scaling coefficient which should scale the loss which gets returned.
        sample_new_memory_batch: Whether a new batch of data should be sampled from the memory
            buffer when the loss is calculated.
        model: The model that is being trained.
        stable_model_update_weight: The weight used in the update of the stable model.
        plastic_model_update_weight:  The weight used in the update of the plastic model.
        stable_model_update_probability:  The probability of updating the stable model at each
            training step.
        plastic_model_update_probability:  The probability of updating the plastic model at each
            training step.
    """

    def __init__(
        self,
        weight: float,
        sample_new_memory_batch: bool,
        model: RenateModule,
        stable_model_update_weight: float,
        plastic_model_update_weight: float,
        stable_model_update_probability: float,
        plastic_model_update_probability: float,
    ) -> None:
        self._stable_model_update_weight = stable_model_update_weight
        self._stable_model_update_weight = stable_model_update_weight
        self._plastic_model_update_weight = plastic_model_update_weight
        self._stable_model_update_probability = stable_model_update_probability
        self._plastic_model_update_probability = plastic_model_update_probability
        self._iteration = 0
        super().__init__(weight=weight, sample_new_memory_batch=sample_new_memory_batch)
        self._plastic_model: RenateModule = copy.deepcopy(model)
        self._stable_model: RenateModule = copy.deepcopy(model)
        self._plastic_model.deregister_hooks()
        self._stable_model.deregister_hooks()

    def _verify_attributes(self) -> None:
        """Verify if attributes have valid values."""
        super()._verify_attributes()
        assert 0.0 <= self._stable_model_update_weight
        assert 0.0 <= self._plastic_model_update_weight
        assert 0.0 <= self._stable_model_update_probability <= 1.0
        assert 0.0 <= self._plastic_model_update_probability <= 1.0
        assert self._plastic_model_update_probability > self._stable_model_update_probability
        assert self._plastic_model_update_weight <= self._stable_model_update_weight

    def _loss(
        self,
        outputs_memory: torch.Tensor,
        batch_memory: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        intermediate_representation_memory: Optional[List[torch.Tensor]],
    ) -> torch.Tensor:
        """Computes the consistency loss with respect to averaged plastic and stable models."""
        (inputs_memory, targets_memory), _ = batch_memory
        with torch.no_grad():
            outputs_plastic = self._plastic_model(inputs_memory)
            outputs_stable = self._stable_model(inputs_memory)
            probs_plastic = F.softmax(outputs_plastic, dim=-1)
            probs_stable = F.softmax(outputs_stable, dim=-1)
            label_mask = F.one_hot(targets_memory, num_classes=outputs_stable.shape[-1]) > 0
            idx = (probs_stable[label_mask] > probs_plastic[label_mask]).unsqueeze(1)
            outputs = torch.where(idx, outputs_stable, outputs_plastic)
        consistency_loss = F.mse_loss(outputs_memory, outputs.detach(), reduction="mean")
        return self.weight * consistency_loss

    @torch.no_grad()
    def _update_model_variables(
        self, model: RenateModule, original_model: RenateModule, weight: float
    ) -> None:
        """Performs exponential moving average on the stored model copies.

        Args:
            model: Whether the plastic or the stable model is updated.
            weight: The minimum weight used in the exponential moving average to update the model.
        """
        alpha = min(1.0 - 1.0 / (self._iteration + 1), weight)
        for ema_p, p in zip(model.parameters(), original_model.parameters()):
            ema_p.data.mul_(alpha).add_(p.data, alpha=1 - alpha)

    def on_train_batch_end(self, model: RenateModule) -> None:
        """Updates the model copies with the current weights,
        given the specified probabilities of update, and increments iteration counter."""
        self._iteration += 1
        if torch.rand(1) < self._plastic_model_update_probability:
            self._update_model_variables(
                self._plastic_model, model, self._plastic_model_update_weight
            )

        if torch.rand(1) < self._stable_model_update_probability:
            self._update_model_variables(
                self._stable_model, model, self._stable_model_update_weight
            )

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Load relevant information from checkpoint."""
        super().on_load_checkpoint(checkpoint)
        self._plastic_model.load_state_dict(checkpoint["component-cls-plastic-model"])
        self._stable_model.load_state_dict(checkpoint["component-cls-stable-model"])
        self._iteration = checkpoint["component-cls-iteration"]

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """Add plastic and stable model to checkpoint."""
        super().on_save_checkpoint(checkpoint)
        checkpoint["component-cls-plastic-model"] = self._plastic_model.state_dict()
        checkpoint["component-cls-stable-model"] = self._stable_model.state_dict()
        checkpoint["component-cls-iteration"] = self._iteration
